// Copyright (c) 2001, 2002 Zope Corporation and Contributors.
//
// All Rights Reserved.
//
// This software is subject to the provisions of the Zope Public License,
// Version 2.0 (ZPL).  A copy of the ZPL should accompany this distribution.
// THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
// WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
// WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
// FOR A PARTICULAR PURPOSE.

using System;
using System.Collections;
using System.Reflection;

namespace Python.Runtime {

    /// <summary>
    /// Manages the tracking of assemblies and namespaces and provides a
    /// simplified internal interface for finding and obtaining Types.
    /// </summary>

    internal class AssemblyManager {

	static Hashtable namespaces;
	static ArrayList assemblies;
	static Hashtable probed;

	private AssemblyManager() {}

	static AssemblyManager() {
	    namespaces = new Hashtable();
	    assemblies = new ArrayList();
	    probed = new Hashtable();

	    AppDomain domain = AppDomain.CurrentDomain;
	    domain.AssemblyLoad += new AssemblyLoadEventHandler(
				       AssemblyLoadHandler
				       );
	
	    Assembly[] items = domain.GetAssemblies();
	    for (int i = 0; i < items.Length; i++) {
		Assembly a = items[i];
		assemblies.Add(a);
		ScanAssembly(a);
	    }
	}

	//===================================================================
	// Event handler for assembly load events. This ensures that we
	// can track all assemblies loaded into the running app domain.
	//===================================================================

	static void AssemblyLoadHandler(Object ob, AssemblyLoadEventArgs args){
	    Assembly assembly = args.LoadedAssembly;
	    assemblies.Add(assembly);
	    ScanAssembly(assembly);
	}

	//===================================================================
	// Loads an assembly from the application directory or the GAC
	// given a simple assembly name. Returns true if an assembly was 
	// loaded, false if the assembly was not found or already loaded.
	//===================================================================

	public static bool LoadAssembly(string name) {
	    Assembly assembly = null;
	    try {
		assembly = Assembly.LoadWithPartialName(name);
	    }
	    catch {
	    }
	    return (assembly != null);
	}

	//===================================================================
	// Given a qualified name of the form A.B.C.D, attempt to load 
	// an assembly named after each of A.B.C.D, A.B.C, A.B, A. This
	// will only actually probe for the assembly once for each unique
	// namespace. Returns true if any assemblies were loaded.
	//===================================================================

	public static bool LoadImplicit(string name) {
	    string[] names = name.Split('.');
	    bool loaded = false;
	    string s = "";
	    for (int i = 0; i < names.Length; i++) {
		s = (i == 0) ? names[0] : s + "." + names[i];
		if (probed[s] == null) {
		    if (LoadAssembly(s)) {
			loaded = true;
		    }
		    probed[s] = 1;
		}
	    }
	    return loaded;
	}

	//===================================================================
	// Scans an assembly for exported namespaces, adding them to the
	// mapping of valid namespaces. Note that for a given namespace
	// a.b.c.d, each of a, a.b, a.b.c and a.b.c.d are considered to 
	// be valid namespaces (to better match Python import semantics).
	//===================================================================

	static void ScanAssembly(Assembly assembly) {
	    Type[] types = assembly.GetTypes();
	    for (int i = 0; i < types.Length; i++) {
		string type_ns = types[i].Namespace;
		if ((type_ns != null) && (!namespaces.ContainsKey(type_ns))) {
		    string[] names = type_ns.Split('.');
		    string s = "";
		    for (int n = 0; n < names.Length; n++) {
			s = (n == 0) ? names[0] : s + "." + names[n];
			if (!namespaces.ContainsKey(s)) {
			    namespaces.Add(s, String.Empty);
			}
		    }
		}
	    }
	}

	//===================================================================
	// Returns true if the given qualified name matches a namespace
	// exported by an assembly loaded in the current app domain.
	//===================================================================

	public static bool IsValidNamespace(string name) {
	    return namespaces.ContainsKey(name);
	}

	//===================================================================
	// Returns the System.Type object for a given qualified name,
	// looking in the currently loaded assemblies for the named
	// type. Returns null if the named type cannot be found.
	//===================================================================

	public static Type LookupType(string qualifiedName) {
	    for (int i = 0; i < assemblies.Count; i++) {
		Assembly assembly = (Assembly)assemblies[i];
		Type type = assembly.GetType(qualifiedName);
		if (type != null) {
		    return type;
		}
	    }
	    return null;
	}

    }


}
