// Copyright (c) 2001, 2002 Zope Corporation and Contributors.
//
// All Rights Reserved.
//
// This software is subject to the provisions of the Zope Public License,
// Version 2.0 (ZPL).  A copy of the ZPL should accompany this distribution.
// THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
// WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
// WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
// FOR A PARTICULAR PURPOSE.

using System;
using System.Collections;
using System.Reflection;
using System.Runtime.InteropServices;

namespace Python.Runtime {

    /// <summary>
    /// Base class for Python types that reflect managed types / classes.
    /// Concrete subclasses include ClassObject and DelegateObject. This
    /// class provides common attributes and common machinery for doing
    /// class initialization (initialization of the class __dict__). The
    /// concrete subclasses provide slot implementations appropriate for
    /// each variety of reflected type.
    /// </summary>

    internal class ClassBase : ManagedType {

	internal Type type;

	internal ClassBase(Type tp) : base() {
	    tpHandle = TypeManager.GetTypeHandle(this, tp);
	    type = tp;
	    InitializeClass();
	}

	internal virtual bool CanSubclass() {
	    return (!this.type.IsEnum);
	}

	//====================================================================
	// Reflected type initialization. This uses reflection to populate the
	// __dict__ of the Python type object with descriptors that reflect
	// the attributes of the reflected type.
	//====================================================================

	protected void InitializeClass() {
	    // todo: name discrimination!
	    Hashtable methods = new Hashtable();
	    IntPtr pp = Runtime._PyObject_GetDictPtr(this.tpHandle);
	    IntPtr op = Marshal.ReadIntPtr(pp);


	    PyDict dict = new PyDict(op);
	    Type type = this.type;

	    ArrayList list;
	    MethodInfo mInfo;
	    String name;
	    Object item;
	    Type tp;

	    ManagedType mt;
	    // may want to use DeclaredOnly here to match Python inheritance.

	    BindingFlags flags = BindingFlags.Static | 
		                 BindingFlags.Instance | 
	                         BindingFlags.Public | 
                                 BindingFlags.NonPublic;

	    MemberInfo[] info = type.GetMembers(flags);

	    int ilen = info.Length;
	    for (int i=0; i < ilen; i++) {

		MemberInfo mi = info[i];


		switch(mi.MemberType) {

	        case MemberTypes.Method:
		    mInfo = (MethodInfo) mi;
		    if (!(mInfo.IsPublic || mInfo.IsFamily || 
			  mInfo.IsFamilyOrAssembly))
			continue;
		    name = mInfo.Name;
		    item = methods[name];
		    if (item == null) {
			item = methods[name] = new ArrayList();
		    }
		    list = (ArrayList) item;
		    list.Add(mInfo);
		    continue;

	        case MemberTypes.Property:
		    PropertyInfo pi = (PropertyInfo) mi;

		    MethodInfo mm = null;
		    try {
			mm = pi.GetGetMethod(true);
			if (mm == null) {
			    mm = pi.GetSetMethod(true);
			}
		    }
		    catch (System.Security.SecurityException) {
			// GetGetMethod may try to get a method protected by
			// StrongNameIdentityPermission - effectively private.
			continue;
		    }

		    if (mm == null) {
			continue;
		    }

		    if (!(mm.IsPublic || mm.IsFamily || mm.IsFamilyOrAssembly))
			continue;
		    mt = new PropertyObject((PropertyInfo) mi);
		    dict[mi.Name] = (PyObject) mt;
		    continue;

	        case MemberTypes.Field:
		    FieldInfo fi = (FieldInfo) mi;
		    if (!(fi.IsPublic || fi.IsFamily || fi.IsFamilyOrAssembly))
			continue;
		    mt = new FieldObject((FieldInfo) mi);
		    dict[mi.Name] = (PyObject) mt;
		    continue;

	        case MemberTypes.Event:
		    mt = new EventObject((EventInfo) mi);
		    dict[mi.Name] = (PyObject) mt;
		    continue;

	        case MemberTypes.NestedType:
		    tp = (Type) mi;
		    if (!(tp.IsNestedPublic || tp.IsNestedFamily || 
			  tp.IsNestedFamORAssem))
			continue;
		    mt = ClassManager.GetClass(tp);
		    dict[mi.Name] = (PyObject) mt;
		    continue;

	        case MemberTypes.TypeInfo:
		    continue;

	        case MemberTypes.Constructor:
		    continue;

	        case MemberTypes.Custom:
		    continue;
		}
	    }

	    IDictionaryEnumerator dictIter = methods.GetEnumerator();

	    while(dictIter.MoveNext()) {
		name = (string) dictIter.Key;
		list = (ArrayList) dictIter.Value;

		MethodInfo[] mlist = (MethodInfo[])list.ToArray(typeof(MethodInfo));

		mt = new MethodObject(name, mlist);
		dict[name] = (PyObject) mt;

	    }
	}


	//====================================================================
	// Standard comparison implementation for instances of reflected types.
	//====================================================================

	[CallConvCdecl()]
	public static int tp_compare(IntPtr ob, IntPtr other) {
	    if (ob == other) {
		return 0;
	    }
	    CLRObject co1 = GetManagedObject(ob) as CLRObject;
	    CLRObject co2 = GetManagedObject(other) as CLRObject;
	    Object o1 = co1.inst;
	    Object o2 = co2.inst;
	    if (Object.Equals(o1, o2)) {
		return 0;
	    }
	    return -1;
	}

	//====================================================================
	// Standard dealloc implementation for instances of reflected types.
	//====================================================================

	[CallConvCdecl()]
	public static void tp_dealloc(IntPtr ob) {
	    ManagedType self = GetManagedObject(ob);
	    //Console.WriteLine("dealloc: {0}", self.GetType());
	    Runtime.PyMem_Free(self.pyHandle);
	    Runtime.Decref(self.tpHandle);
	    self.gcHandle.Free();
	}


    }	

}
