// Copyright (c) 2001, 2002 Zope Corporation and Contributors.
//
// All Rights Reserved.
//
// This software is subject to the provisions of the Zope Public License,
// Version 2.0 (ZPL).  A copy of the ZPL should accompany this distribution.
// THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
// WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
// WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
// FOR A PARTICULAR PURPOSE.

using System;
using System.Reflection;
using System.Runtime.InteropServices;
using System.Security;

namespace Python.Runtime {

    //========================================================================
    // Performs data conversions between managed types and Python types.
    //========================================================================

    [SuppressUnmanagedCodeSecurityAttribute()]

    internal class Converter {

	private Converter() {}

	static Type objectType;
	static Type flagsType;

	static Converter () {
	    objectType = typeof(Object);
	    flagsType = typeof(FlagsAttribute);
	}


//  	internal static IntPtr ToPython(Object value) {
//  	    if (value == null) {
//  		IntPtr result = Runtime.PyNone;
//  		Runtime.Incref(result);
//  		return result;
//  	    }
//  	    return ToPython(value, value.GetType());
//  	}


	//====================================================================
	// Return a Python object for the given native object, converting
	// basic types (string, int, etc.) into equivalent Python objects.
	// This always returns a new reference. Note that the System.Decimal 
	// type has no Python equivalent and converts to a managed instance.
	//====================================================================

	internal static IntPtr ToPython(Object value, Type type) {
	    IntPtr result = IntPtr.Zero;

	    if (value == null) {
		result = Runtime.PyNone;
		Runtime.Incref(result);
		return result;
	    }

	    // Special case: if the input conversion type is 'Object', we
	    // convert to the actual runtime type of the input value.

	    if (type == objectType) {
		type = value.GetType();
	    }

	    TypeCode tc = Type.GetTypeCode(type);

	    switch(tc) {

	    case TypeCode.Object:
		result = CLRObject.GetInstHandle(value, type);
		Runtime.Incref(result);
		return result;

	    case TypeCode.String:
		return Runtime.PyUnicode_FromString((string)value);

	    case TypeCode.Int32:
		return Runtime.PyInt_FromLong((int)value);

	    case TypeCode.Boolean:
		int i = (bool) value ? 1 : 0;
		return Runtime.PyInt_FromLong(i);

	    case TypeCode.Byte:
		return Runtime.PyInt_FromLong((int)((byte)value));

	    case TypeCode.Char:
		return Runtime.PyUnicode_FromOrdinal((int)((char)value));
	      
	    case TypeCode.Int16:
		return Runtime.PyInt_FromLong((int)((short)value));

	    case TypeCode.Int64:
		return Runtime.PyLong_FromLongLong((long)value);

	    case TypeCode.Single:
	        string ss = ((float)value).ToString();
		IntPtr ps = Runtime.PyString_FromString(ss);
	        IntPtr op = Runtime.PyFloat_FromString(ps, IntPtr.Zero);
		Runtime.Decref(ps);
		return op;

	    case TypeCode.Double:
		return Runtime.PyFloat_FromDouble((double)value);

	    case TypeCode.SByte:
		return Runtime.PyInt_FromLong((int)((sbyte)value));

	    case TypeCode.UInt16:
		return Runtime.PyInt_FromLong((int)((ushort)value));

	    case TypeCode.UInt32:
		return Runtime.PyLong_FromUnsignedLong((uint)value);

	    case TypeCode.UInt64:
		return Runtime.PyLong_FromUnsignedLongLong((ulong)value);

	    default:
		result = CLRObject.GetInstHandle(value, type);
		Runtime.Incref(result);
		return result;
	    }

	}

	//====================================================================
	// Return a managed object for the given Python object, converting
	// basic types (string, int, etc.) into required managed objects.
	//====================================================================

	internal static bool ToManaged(IntPtr value, Type obType, 
				      out Object result, bool setError) {

	    // Common case: if the Python value is a wrapped managed object
	    // instance, just return the wrapped object.

	    ManagedType mt = ManagedType.GetManagedObject(value);
	    result = null;

	    if (mt != null) {
		if (mt is CLRObject) {
		    result = ((CLRObject)mt).inst;
		    return true;
		}
		if (mt is ClassBase) {
		    result = ((ClassBase)mt).type;
		    return true;
		}
		// shouldnt happen
		return false;
	    }

	    if (obType.IsEnum) {
		return ToEnum(value, obType, out result, setError);
	    }

	    if (value == Runtime.PyNone && !obType.IsPrimitive) {
		result = null;
		return true;
	    }

	    return ToPrimitive(value, obType, out result, setError);

	}

	//====================================================================
	// Convert a Python value to an instance of a primitive managed type.
	//====================================================================

	static bool ToPrimitive(IntPtr value, Type obType, out Object result, 
				bool setError) {

	    PyObject overflow = Exceptions.OverflowError;
	    TypeCode tc = Type.GetTypeCode(obType);
	    result = null;
	    IntPtr op;
	    int ival;

	    switch(tc) {

	    case TypeCode.String:
		string st = Runtime.GetManagedString(value);
		if (st == null) {
		    goto type_error;
		}
		result = st;
		return true;

	    case TypeCode.Int32:
		// Trickery to support 64-bit platforms.
	        if (IntPtr.Size == 4) {
		    op = Runtime.PyNumber_Int(value);
		    if (op == IntPtr.Zero) {
		        if (Exceptions.ExceptionMatches(overflow)) {
			    goto overflow;
		      }
		      goto type_error;
		    }
		    ival = (int)Runtime.PyInt_AsLong(op);
		    Runtime.Decref(op);
		    result = ival;
		    return true;
		}
		else {
		    op = Runtime.PyNumber_Long(value);
		    if (op == IntPtr.Zero) {
		        if (Exceptions.ExceptionMatches(overflow)) {
			    goto overflow;
			}
			goto type_error;
		    }
		    long ll = (long)Runtime.PyLong_AsLongLong(op);
		    Runtime.Decref(op);
		    if ((ll == -1) && Exceptions.ErrorOccurred()) {
		        goto overflow;
		    }
		    if (ll > Int32.MaxValue || ll < Int32.MinValue) {
		        goto overflow;
		    }
		    result = (int)ll;
		    return true;
		}

	    case TypeCode.Boolean:
		result = (Runtime.PyObject_IsTrue(value) != 0);
		return true;

	    case TypeCode.Byte:
		op = Runtime.PyNumber_Int(value);
		if (op == IntPtr.Zero) {
		    if (Exceptions.ExceptionMatches(overflow)) {
			goto overflow;
		    }
		    goto type_error;
		}
		ival = (int) Runtime.PyInt_AsLong(op);
		Runtime.Decref(op);

		if (ival > Byte.MaxValue || ival < Byte.MinValue) {
		    goto overflow;
		}
		byte b = (byte) ival;
		result = b;
		return true;

	    case TypeCode.SByte:
		op = Runtime.PyNumber_Int(value);
		if (op == IntPtr.Zero) {
		    if (Exceptions.ExceptionMatches(overflow)) {
			goto overflow;
		    }
		    goto type_error;
		}
		ival = (int) Runtime.PyInt_AsLong(op);
		Runtime.Decref(op);

		if (ival > SByte.MaxValue || ival < SByte.MinValue) {
		    goto overflow;
		}
		sbyte sb = (sbyte) ival;
		result = sb;
		return true;

	    case TypeCode.Char:

		if (Runtime.PyObject_TypeCheck(value, Runtime.PyStringType)) {
		    if (Runtime.PyString_Size(value) == 1) {
			op = Runtime.PyString_AS_STRING(value);
			result = (char)Marshal.ReadByte(op);
			return true;
		    }
		    goto type_error;
		}

		else if (Runtime.PyObject_TypeCheck(value, 
				 Runtime.PyUnicodeType)) {
		    if (Runtime.PyUnicode_GetSize(value) == 1) {
			op = Runtime.PyUnicode_AS_UNICODE(value);
			result = (char)Marshal.ReadInt16(op);
			return true;
		    }
		    goto type_error;
		}

		op = Runtime.PyNumber_Int(value);
		if (op == IntPtr.Zero) {
		    goto type_error;
		}
		ival = Runtime.PyInt_AsLong(op);
		if (ival > Char.MaxValue || ival < Char.MinValue) {
		    goto overflow;
		}
		Runtime.Decref(op);
		result = (char)ival;
		return true;

	    case TypeCode.Int16:
		op = Runtime.PyNumber_Int(value);
		if (op == IntPtr.Zero) {
		    if (Exceptions.ExceptionMatches(overflow)) {
			goto overflow;
		    }
		    goto type_error;
		}
		ival = (int) Runtime.PyInt_AsLong(op);
		Runtime.Decref(op);
		if (ival > Int16.MaxValue || ival < Int16.MinValue) {
		    goto overflow;
		}
		short s = (short) ival;
		result = s;
		return true;

	    case TypeCode.Int64:
		op = Runtime.PyNumber_Long(value);
		if (op == IntPtr.Zero) {
		    if (Exceptions.ExceptionMatches(overflow)) {
			goto overflow;
		    }
		    goto type_error;
		}
		long l = (long)Runtime.PyLong_AsLongLong(op);
		Runtime.Decref(op);
		if ((l == -1) && Exceptions.ErrorOccurred()) {
		    goto overflow;
		}
		result = l;
		return true;

	    case TypeCode.UInt16:
		op = Runtime.PyNumber_Int(value);
		if (op == IntPtr.Zero) {
		    if (Exceptions.ExceptionMatches(overflow)) {
			goto overflow;
		    }
		    goto type_error;
		}
		ival = (int) Runtime.PyInt_AsLong(op);
		Runtime.Decref(op);
		if (ival > UInt16.MaxValue || ival < UInt16.MinValue) {
		    goto overflow;
		}
		ushort us = (ushort) ival;
		result = us;
		return true;

	    case TypeCode.UInt32:
		op = Runtime.PyNumber_Long(value);
		if (op == IntPtr.Zero) {
		    if (Exceptions.ExceptionMatches(overflow)) {
			goto overflow;
		    }
		    goto type_error;
		}
		uint ui = (uint)Runtime.PyLong_AsUnsignedLong(op);
		Runtime.Decref(op);
		if (Exceptions.ErrorOccurred()) {
		    goto overflow;
		}
		result = ui;
		return true;

	    case TypeCode.UInt64:
		op = Runtime.PyNumber_Long(value);
		if (op == IntPtr.Zero) {
		    if (Exceptions.ExceptionMatches(overflow)) {
			goto overflow;
		    }
		    goto type_error;
		}
		ulong ul = (ulong)Runtime.PyLong_AsUnsignedLongLong(op);
		Runtime.Decref(op);
		if (Exceptions.ErrorOccurred()) {
		    goto overflow;
		}
		result = ul;
		return true;


	    case TypeCode.Single:
		op = Runtime.PyNumber_Float(value);
		if (op == IntPtr.Zero) {
		    if (Exceptions.ExceptionMatches(overflow)) {
			goto overflow;
		    }
		    goto type_error;
		}
		double dd = Runtime.PyFloat_AsDouble(value);
		if (dd > Single.MaxValue || dd < Single.MinValue) {
		    goto overflow;
		}
		result = (float)dd;
		return true;

	    case TypeCode.Double:
		op = Runtime.PyNumber_Float(value);
		if (op == IntPtr.Zero) {
		    goto type_error;
		}
		double d = Runtime.PyFloat_AsDouble(op);
		Runtime.Decref(op);
		if (d > Double.MaxValue || d < Double.MinValue) {
		    goto overflow;
		}
		result = d;
		return true;

	    }


	type_error:

	    if (setError) {
		string format = "'{0}' value cannot be converted to {1}";
		string tpName = Runtime.PyObject_GetTypeName(value);
		string error = String.Format(format, tpName, obType);
		Exceptions.SetError(Exceptions.TypeError, error);
	    }

	    return false;

	overflow:

	    if (setError) {
		string error = "value too large to convert";
		Exceptions.SetError(Exceptions.OverflowError, error);
	    }

	    return false;

	}

	//====================================================================
	// Convert a Python value to a correctly typed managed enum instance.
	//====================================================================

	static bool ToEnum(IntPtr value, Type obType, out Object result, 
			   bool setError) {

	    Type etype = Enum.GetUnderlyingType(obType);
	    result = null;

	    if (!ToPrimitive(value, etype, out result, setError)) {
		return false;
	    }

	    if (Enum.IsDefined(obType, result)) {
		result = Enum.ToObject(obType, result);		
		return true;
	    }

	    if (obType.GetCustomAttributes(flagsType, true).Length > 0) {
		result = Enum.ToObject(obType, result);
		return true;
	    }

	    if (setError) {
		string error = "invalid enumeration value";
		Exceptions.SetError(Exceptions.ValueError, error);
	    }

	    return false;

	}

    }


}
