// Copyright (c) 2003 Zope Corporation and Contributors.
//
// All Rights Reserved.
//
// This software is subject to the provisions of the Zope Public License,
// Version 2.0 (ZPL).  A copy of the ZPL should accompany this distribution.
// THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
// WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
// WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
// FOR A PARTICULAR PURPOSE.

using System;
using System.Reflection;

namespace Python.Runtime {

    /// <summary>
    /// Managed class that provides the implementation for reflected delegate
    /// types. Delegates are represented in Python by generated type objects. 
    /// Each of those type objects is associated an instance of this class, 
    /// which provides its implementation.
    /// </summary>

    internal class DelegateObject : ClassBase {

	MethodBinder binder;

	internal DelegateObject(Type tp) : base(tp) {
	    binder = new MethodBinder(tp.GetMethod("Invoke"));
	}

	//====================================================================
	// Given a PyObject pointer to an instance of a delegate type, return
	// the true managed delegate the Python object represents (or null).
	//====================================================================

	private static Delegate GetTrueDelegate(IntPtr op) {
	    CLRObject o = GetManagedObject(op) as CLRObject;
	    if (o != null) {
		Delegate d = o.inst as Delegate;
		return d;
	    }
	    return null;
	}

	internal override bool CanSubclass() {
	    return false;
	}

	//====================================================================
	// DelegateObject __new__ implementation. The result of this is a new
	// PyObject whose type is DelegateObject and whose ob_data is a handle
	// to an actual delegate instance. The method wrapped by the actual
	// delegate instance belongs to an object generated to relay the call
	// to the Python callable passed in.
	//====================================================================

	[CallConvCdecl()]
	public static IntPtr tp_new(IntPtr tp, IntPtr args, IntPtr kw) {
	    DelegateObject self = (DelegateObject)GetManagedObject(tp);

	    if (Runtime.PyTuple_Size(args) != 1) {
		Exceptions.SetError(Exceptions.TypeError,
				    "class takes exactly one argument"
				    );
		return IntPtr.Zero;
	    }

	    IntPtr method = Runtime.PyTuple_GetItem(args, 0);

	    if (Runtime.PyCallable_Check(method) != 1) {
		Exceptions.SetError(Exceptions.TypeError,
				    "argument must be callable"
				    );
		return IntPtr.Zero;
	    }

	    Delegate d = DelegateManager.GetDelegate(self.type, method);
	    IntPtr op = CLRObject.GetInstHandle(d, self.Handle);
	    Runtime.Incref(op);
	    return op;
	}

	//====================================================================
	// Implements __init__ for reflected delegate types.
	//====================================================================

	[CallConvCdecl()]
	public static int tp_init(IntPtr ob, IntPtr args, IntPtr kw) {
	    return 0;
	}

	//====================================================================
	// Implements __call__ for reflected delegate types.
	//====================================================================

	[CallConvCdecl()]
	public static IntPtr tp_call(IntPtr ob, IntPtr args, IntPtr kw) {
	    // todo: add fast type check!
	    IntPtr pytype = Runtime.PyObject_Type(ob);
	    DelegateObject self = (DelegateObject)GetManagedObject(pytype);
	    CLRObject o = GetManagedObject(ob) as CLRObject;

	    if (o == null) {
		Exceptions.SetError(Exceptions.TypeError, "invalid argument");
		return IntPtr.Zero;
	    }
	    
	    Delegate d = o.inst as Delegate;

	    if (d == null) {
		Exceptions.SetError(Exceptions.TypeError, "invalid argument");
		return IntPtr.Zero;
	    }

	    return self.binder.Invoke(ob, args, kw);
	}

	//====================================================================
	// Implements __cmp__ for reflected delegate types.
	//====================================================================

	[CallConvCdecl()]
	public static new int tp_compare(IntPtr ob, IntPtr other) {
	    Delegate d1 = GetTrueDelegate(ob);
	    Delegate d2 = GetTrueDelegate(other);
	    if (d1 == d2) {
		return 0;
	    }
	    return -1;
	}


	/*
	public override IntPtr ClassRepr() {
	    string s = String.Format("<delegate '{0}'>", this.type.FullName);
	    return Runtime.PyString_FromStringAndSize(s, s.Length);
	}
	*/

    }	

}
