// Copyright (c) 2001, 2002 Zope Corporation and Contributors.
//
// All Rights Reserved.
//
// This software is subject to the provisions of the Zope Public License,
// Version 2.0 (ZPL).  A copy of the ZPL should accompany this distribution.
// THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
// WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
// WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
// FOR A PARTICULAR PURPOSE.

using System;
using System.Collections;
using System.Reflection;

namespace Python.Runtime {

    //========================================================================
    // Implements a Python descriptor type that provides access to CLR events.
    //========================================================================

    internal class EventObject : ExtensionType {

	static Hashtable registry;

	internal string name;
	EventBinding unbound;
	EventInfo info;

	public EventObject(EventInfo info) : base() {
	    this.name = info.Name;
	    this.info = info;
	}

	static EventObject() {
	    registry = new Hashtable();
	}



	public int AddEventHandler(IntPtr target, IntPtr handler) {
	    Object obj = null;
	    if (target != IntPtr.Zero) {
		CLRObject co = (CLRObject)ManagedType.GetManagedObject(target);
		obj = co.inst;
	    }

	    //try {
		Type htype = this.info.EventHandlerType;
		Delegate d = DelegateManager.GetDelegate(htype, handler);
		this.info.AddEventHandler(obj, d);
//  	    }
//  	    catch(Exception e) {
//  		Exceptions.SetError(Exceptions.TypeError, e.Message);
//  		return -1;
//  	    }

	    return 0;
	}

	public int RemoveEventHandler(IntPtr target, IntPtr arg) {
	    return -1;
	}

	public IntPtr Invoke(IntPtr target, IntPtr args, IntPtr kw) {
	    MethodInfo method = info.GetRaiseMethod();
	    MethodBinder binder = new MethodBinder(method);
	    return binder.Invoke(target, args, kw);
	}

	//====================================================================
	// Descriptor __get__ implementation. A getattr on an event returns
	// a "bound" event that keeps a reference to the object instance,
	// making events first-class Python objects similar to methods.
	//====================================================================

	[CallConvCdecl()]
	public static IntPtr tp_descr_get(IntPtr ds, IntPtr ob, IntPtr tp) {
	    EventObject self = (EventObject)GetManagedObject(ds);
	    EventBinding binding;

	    // If the event is accessed through its type (rather than via
	    // an instance) we return an 'unbound' EventBinding that will
	    // cached for future accesses through the type.

	    if (ob == IntPtr.Zero) {
		if (self.unbound == null) {
		    self.unbound = new EventBinding(self, IntPtr.Zero);
		}
		binding = self.unbound;
		Runtime.Incref(binding.Handle);
		return binding.Handle;
	    }

	    if (Runtime.PyObject_IsInstance(ob, tp) == 0) {
		Exceptions.SetError(Exceptions.TypeError, 
				    "invalid descriptor argument"
				    );
		return IntPtr.Zero;
	    }

	    binding = new EventBinding(self, ob);
	    return binding.Handle;
	}

	//====================================================================
	// Descriptor __set__ implementation. This actually never allows you
	// to set anything; it exists solely to support the '+=' spelling of
	// event handler registration. The reason is that given code like:
	// 'ob.SomeEvent += method', Python will attempt to set the attribute
	// SomeEvent on ob to the result of the '+=' operation.
	//====================================================================

	[CallConvCdecl()]
	public static int tp_descr_set(IntPtr ds, IntPtr ob, IntPtr val) {
	    EventObject self = (EventObject)GetManagedObject(ds);
	    ManagedType m = GetManagedObject(val);
	    // fix to make this a simple type check...
	    if (m is EventBinding) {
		return 0;
	    }
	    string message = "cannot set event attributes";
	    Exceptions.SetError(Exceptions.TypeError, message);
	    return -1;
	}


	//====================================================================
	// Descriptor __repr__ implementation.
	//====================================================================

	[CallConvCdecl()]
	public static IntPtr tp_repr(IntPtr ob) {
	    EventObject self = (EventObject)GetManagedObject(ob);
	    string s = String.Format("<event '{0}'>", self.info.Name);
	    return Runtime.PyString_FromStringAndSize(s, s.Length);
	}

	//====================================================================
	// Descriptor dealloc implementation.
	//====================================================================

	[CallConvCdecl()]
	public static new void tp_dealloc(IntPtr ob) {
	    EventObject self = (EventObject)GetManagedObject(ob);
	    if (self.unbound != null) {
		Runtime.Decref(self.unbound.Handle);
	    }
	    ExtensionType.FinalizeObject(self);
	}

    }


}
