// Copyright (c) 2001, 2002 Zope Corporation and Contributors.
//
// All Rights Reserved.
//
// This software is subject to the provisions of the Zope Public License,
// Version 2.0 (ZPL).  A copy of the ZPL should accompany this distribution.
// THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
// WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
// WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
// FOR A PARTICULAR PURPOSE.

using System;
using System.Runtime.InteropServices;
using System.Collections;
using System.Reflection;

namespace Python.Runtime {

    //========================================================================
    // The managed metatype. This object implements the type of all reflected
    // types. It also provides support for single-inheritance from reflected
    // managed types.
    //========================================================================
    
    internal class MetaType : ManagedType {

	static Type delegateType;
	static IntPtr PyCLRMetaType;
	static IntPtr stdflags;
	static ArrayList dList;
	static Hashtable cache;
	static int obSize;
	static int tpSize;

	static MetaType() {
	    obSize = Marshal.SizeOf(new PyObjectHead());
	    tpSize = Marshal.SizeOf(new PyTypeObject());
	    cache = new Hashtable();
	    dList = new ArrayList();
	    stdflags = (IntPtr) (PyTypeFlags.Default | 
				 PyTypeFlags.HeapType |
				 PyTypeFlags.Managed);
	}

	//====================================================================
	// Metatype initialization. This bootstraps the CLR metatype to life.
	//====================================================================

	public static IntPtr Initialize() {

	    Type implType = typeof(MetaType);

	    PyTypeObject pyTypeObj = new PyTypeObject();
	    pyTypeObj.tp_name = Marshal.StringToHGlobalAnsi(implType.Name);
	    pyTypeObj.ob_type = Runtime.PyTypeType;
	    pyTypeObj.tp_base = Runtime.PyTypeType;

	    int typeSize = Marshal.SizeOf(typeof(PyTypeObject));
	    pyTypeObj.tp_basicsize = (IntPtr)typeSize;
	    pyTypeObj.ob_refcnt = (IntPtr) 1;
	    pyTypeObj.tp_flags = (IntPtr) (PyTypeFlags.Default | 
					   PyTypeFlags.HeapType |
					   PyTypeFlags.Managed);

	    IntPtr pThunk = Marshal.AllocHGlobal(IntPtr.Size);
	    BindingFlags flags = BindingFlags.Public | BindingFlags.Static;
	    Type th = typeof(CallbackThunks);
	    Type py = pyTypeObj.GetType();
	    Type tp = implType;

	    while (tp != null) {
		foreach (MethodInfo method in tp.GetMethods(flags)) {
		    string methodName = method.Name;
		    if (! methodName.StartsWith("tp_") ) {
			continue;
		    }

		    FieldInfo slot = py.GetField(methodName);
		    if (slot != null && 
			(IntPtr)slot.GetValue(pyTypeObj) == IntPtr.Zero) {

			Type dt = th.GetNestedType(methodName);
			if (dt != null) {
			    Delegate d = Delegate.CreateDelegate(dt, method);
			    CallbackThunk cb = new CallbackThunk(d);

			    Marshal.StructureToPtr(cb, pThunk, false);
			    IntPtr fp = Marshal.ReadIntPtr(pThunk);
			    slot.SetValue(pyTypeObj, fp);
			    dList.Add(d);
			}
		    }
		}
		tp = tp.BaseType;
	    }

	    Marshal.FreeHGlobal(pThunk);

	    // Now blit the PyTypeObject instance to the unmanaged heap. This
	    // bit lets us avoid pinning objects, which would disrupt the GC.

	    IntPtr typePtr = Runtime.PyMem_Malloc(tpSize);
	    GCHandle gch = GCHandle.Alloc(pyTypeObj, GCHandleType.Pinned);
	    Marshal.StructureToPtr(pyTypeObj, typePtr, false);
	    gch.Free();
	    Runtime.PyType_Ready(typePtr);


	    PyCLRMetaType = typePtr;
	    return typePtr;
	}



	//====================================================================
	// Type __setattr__ implementation for reflected types. Note that this
	// is slightly different than the standard setattr implementation for
	// the normal Python metatype (PyTypeType). We need to look first in
	// the type object of a reflected type for a descriptor in order to 
	// support the right setattr behavior for static fields and properties.
	//====================================================================

	[CallConvCdecl()]
	public static int tp_setattro(IntPtr tp, IntPtr name, IntPtr value) {
	    IntPtr descr = Runtime._PyType_Lookup(tp, name);

	    if (descr != IntPtr.Zero) {
		IntPtr dt = Runtime.PyObject_Type(descr);
		IntPtr fp = Marshal.ReadIntPtr(dt, (35 * IntPtr.Size));
		Runtime.Decref(dt);
		if (fp != IntPtr.Zero) {
		    return NativeCall.Impl.Int_Call_3(fp, descr, name, value);
		}
		Exceptions.SetError(Exceptions.AttributeError,
				    "attribute is read-only");
		return -1;
	    }
	    
	    if (Runtime.PyObject_GenericSetAttr(tp, name, value) < 0) {
		return -1;
	    }

	    return 0;
	}


	//====================================================================
	// Metatype __new__ implementation. This is called to create a new 
	// class / type when a reflected class is subclassed. 
	//====================================================================

	[CallConvCdecl()]
	public static IntPtr tp_new(IntPtr tp, IntPtr args, IntPtr kw) {

	    // check for non-subclassable type here.

	    int len = Runtime.PyTuple_Size(args);
	    if (len < 3) {
		Exceptions.SetError(Exceptions.TypeError, 
				    "invalid argument list"
				    );
		return IntPtr.Zero;
	    }

	    IntPtr name = Runtime.PyTuple_GetItem(args, 0);
	    IntPtr bases = Runtime.PyTuple_GetItem(args, 1);
	    IntPtr dict = Runtime.PyTuple_GetItem(args, 2);

	    // We do not support multiple inheritance, so the bases argument
	    // should be a 1-item tuple containing the type we are subtyping.
	    // That type must itself have a managed implementation. We check
	    // that by making sure its metatype is the CLR metatype.

	    if (Runtime.PyTuple_Size(bases) != 1) {
		Exceptions.SetError(Exceptions.TypeError, 
		"cannot use multiple inheritance with managed classes"
				    );
		return IntPtr.Zero;
	    }

	    IntPtr basetype = Runtime.PyTuple_GetItem(bases, 0);
	    if (Runtime.PyObject_Type(basetype) != PyCLRMetaType) {
		Exceptions.SetError(Exceptions.TypeError, 
		"metatype is not managed metatype - should not happen!"
				    );
		return IntPtr.Zero;

	    }

	    // Ensure that the reflected type is appropriate for subclassing,
	    // disallowing subclassing of delegates, enums and array types.

	    ClassBase cb = (ClassBase)GetManagedObject(basetype);
	    if (! cb.CanSubclass() ) {
		Exceptions.SetError(Exceptions.TypeError,
		"delegates, enums and array types cannot be subclassed"
				    );
		return IntPtr.Zero;
	    }


	    IntPtr slots = Runtime.PyDict_GetItemString(dict, "__slots__");
	    if (slots != IntPtr.Zero) {
		Exceptions.SetError(Exceptions.TypeError, 
		"subclasses of managed classes do not support __slots__"
				    );
		return IntPtr.Zero;

	    }

	    dict = Runtime.PyDict_Copy(dict);

	    // todo: set module

	    // todo: set doc

	    // todo: kill __new__ or error

	    // Create the subtype. Note that this does not go through the
	    // standard protocols (tp_alloc, etc.) because we allow only
	    // single inheritance and don't interoperate with other types.

	    PyTypeObject sub = new PyTypeObject();
	    string tp_name = Runtime.PyString_AsString(name);
	    sub.tp_name = Marshal.StringToHGlobalAnsi(tp_name);
	    sub.ob_type = PyCLRMetaType;
	    sub.ob_refcnt = (IntPtr) 1;
	    sub.tp_flags = stdflags;
	    sub.tp_basicsize = (IntPtr)tpSize;

	    sub.tp_base = basetype;
	    Runtime.Incref(basetype);

	    sub.tp_bases = bases;
	    Runtime.Incref(bases);

	    sub.tp_dict = dict;

	    // get/set attr copy?

	    // copy hidden obj
  	    sub.ob_size = Marshal.ReadIntPtr(basetype, 2 * IntPtr.Size);

	    IntPtr op = Runtime.PyMem_Malloc(tpSize);
	    GCHandle gc = GCHandle.Alloc(sub, GCHandleType.Pinned);
	    Marshal.StructureToPtr(sub, op, false);
	    gc.Free();

	    TypeManager.FixupOffsets(op);

	    if (Runtime.PyType_Ready(op) < 0) {
		Runtime.Decref(op);
		return IntPtr.Zero;
	    }

	    return op;
	}


	//====================================================================
	// Dealloc implementation. This is called when a subclass generated
	// by this metatype is no longer referenced in Python.
	//====================================================================

	[CallConvCdecl()]
	public static void tp_dealloc(IntPtr tp) {
	    // Free tp->tp_name, dealloc tp_dict, free type memory.
	    IntPtr op = Marshal.ReadIntPtr(tp, 3 * IntPtr.Size);
	    Marshal.FreeHGlobal(op);
	    op = Marshal.ReadIntPtr(tp, 33 * IntPtr.Size);
	    Runtime.Decref(op);
	    Runtime.PyMem_Free(tp);
	}

	/*
	[CallConvCdecl()]
	public static IntPtr tp_repr(IntPtr ob) {
	    ClassBase self = (ClassBase)GetManagedObject(ob);
	    return self.ClassRepr();
	}
	*/

    }


}
