// Copyright (c) 2003 Zope Corporation and Contributors.
//
// All Rights Reserved.
//
// This software is subject to the provisions of the Zope Public License,
// Version 2.0 (ZPL).  A copy of the ZPL should accompany this distribution.
// THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
// WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
// WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
// FOR A PARTICULAR PURPOSE.

using System;
using System.Collections;
using System.Reflection;

namespace Python.Runtime {

    //========================================================================
    // A MethodBinder encapsulates information about a (possibly overloaded) 
    // managed method, and is responsible for selecting the right method given
    // a set of Python arguments. This is also used as a base class for the
    // ConstructorBinder, a minor variation used to invoke constructors.
    //========================================================================

    internal class MethodBinder {

	MethodInfo info;
	ArrayList list;

	internal MethodBinder () {
	    this.info = null;
	    this.list = new ArrayList();
	}
	
	internal MethodBinder(MethodInfo mi) : base () {
	    this.info = null;
	    this.list = new ArrayList();
	    this.list.Add(mi);
	}

	internal void AddMethod(MethodBase m) {
	    // if more than one, use complex ds, else use info
	    this.list.Add(m);
	}

	internal Binding Bind(IntPtr inst, IntPtr args, IntPtr kw) {
	    // loop to find match, return invoker w/ or /wo error
	    int nargs = Runtime.PyTuple_Size(args);
	    MethodBase[] methods = (MethodBase[])list.ToArray(typeof(MethodBase));
	    
	    for (int i = 0; i < methods.Length; i++) {
		MethodBase mi = methods[i];
		ParameterInfo[] pi = mi.GetParameters();
		int count = pi.Length;

		if ( nargs == count ) {
		    Object[] margs = new Object[count];

		    for (int n = 0; n < count; n++) {
			IntPtr op = Runtime.PyTuple_GetItem(args, n);
			Type type = pi[n].ParameterType;
			Object arg;
			if (!Converter.ToManaged(op, type, out arg, false)) {
			    margs = null;
			    break;
			}
			margs[n] = arg;
		    }
		    
		    if (margs == null) {
			continue;
		    }

		    Object target = null;
		    if ((!mi.IsStatic) && (inst != IntPtr.Zero)) {
			CLRObject co = (CLRObject)ManagedType.GetManagedObject(
                                                  inst
						  );
			target = co.inst;
		    }

		    return new Binding(mi, target, margs);
		}

	    }

	    return null;
	}

	internal virtual IntPtr Invoke(IntPtr inst, IntPtr args, IntPtr kw) {
	    Binding binding = this.Bind(inst, args, kw);
	    Object result;

	    if (binding == null) {
		Exceptions.SetError(Exceptions.TypeError, 
				    "no method matches given arguments"
				    );
		return IntPtr.Zero;
	    }

	    try {
		result = binding.info.Invoke(binding.inst, 
					     BindingFlags.Default, 
					     null, 
					     binding.args, 
					     null);
	    }
	    catch (Exception e) {
		if (e.InnerException != null) {
		    e = e.InnerException;
		}
		Exceptions.SetError(e);
		return IntPtr.Zero;
	    }

	    MethodInfo mi = (MethodInfo)binding.info;
	    return Converter.ToPython(result, mi.ReturnType);
	}

    }


    //========================================================================
    // A Binding is a utility instance that bundles together a MethodInfo
    // representing a method to call, a (possibly null) target instance for
    // the call, and the arguments for the call (all as managed values).
    //========================================================================

    internal class Binding {

	public MethodBase info;
	public Object[] args;
	public Object inst;

	internal Binding(MethodBase info, Object inst, Object[] args) {
	    this.info = info;
	    this.inst = inst;
	    this.args = args;
	}

    }

}
