// Copyright (c) 2001, 2002 Zope Corporation and Contributors.
//
// All Rights Reserved.
//
// This software is subject to the provisions of the Zope Public License,
// Version 2.0 (ZPL).  A copy of the ZPL should accompany this distribution.
// THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
// WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
// WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
// FOR A PARTICULAR PURPOSE.

using System;

namespace Python.Runtime {

    /// <summary>
    /// Represents an arbitrary Python object.
    /// </summary>

    public class PyObject : IDisposable {

	// Reference to the native PyObject pointer.
	internal protected IntPtr handle = IntPtr.Zero;

	/// <summary>
	/// PyObject Constructor
	/// </summary>
	///
	/// <remarks>
	/// Creates a new PyObject from an IntPtr object reference.
	/// </remarks>
	///
	/// <param name='pointer'>
	/// The native PyObject pointer.
	/// </param>
	///
	/// <param name='borrowed'>
	/// A boolean that specifies whether the PyObject pointer is a
	/// borrowed reference. This ensures that the reference count
	/// of the underlying object is handled correctly when the
	/// PyObject instance is garbage collected.
	/// </param>


	public PyObject(IntPtr pointer) {
	    handle = pointer;
	    Runtime.Incref(pointer);
	}

	/// <summary>
	/// Dispose Method
	/// </summary>
	///
	/// <remarks>
	/// The Dispose method provides a way to explicitly release the 
	/// Python object represented by a PyObject instance. Normally
	/// the .NET garbage collector will ensure that Python object
	/// references are managed correctly, but in some cases it is
	/// desirable to release an object reference as soon as possible
	/// (to free an object that holds limited resources, for example)
	/// without waiting for the garbage collector to get around to it.
	/// </remarks>

	public void Dispose() {
	    Console.WriteLine("dispose called: {0}", this.ToString());
	    GC.SuppressFinalize(this);
	    //Runtime.Decref(handle);
	    handle = IntPtr.Zero;
	}

	/// <summary>
	/// Finalize Method
	/// </summary>
	///
	/// <remarks>
	/// Provides support for automatic reference management for 
	/// PyObject instances. This method is called by the runtime
	/// and should not be called by application code.
	/// </remarks>

	public void Finalize() {
	    Dispose();
	}


	// Object Properties

	public PyObject PythonType {
	    get {
		return new PyObject(Runtime.PyObject_Type(handle));
	    }
	}

	public int RefCount {
	    get {
		return Runtime.GetRefCount(handle);
	    }
	}

	/// <summary>
	/// Handle Property
	/// </summary>
	///
	/// <remarks>
	/// Gets the native handle of the underlying Python object.
	/// </remarks>

	public IntPtr Handle {
	    get {
		return handle;
	    }
	}

	public virtual int Length {
	    get {
		int result = Runtime.PyObject_Size(handle);
		if (result < 0) {
		    Runtime.PyErr_Clear();
		    return 0;
		}
		return result;
	    }
	}


	public void Incref() {
	    Runtime.Incref(handle);
	}

	public void Decref() {
	    Runtime.Decref(handle);
	}

	public bool HasAttr(string name) {
	    return (Runtime.PyObject_HasAttrString(handle, name) != 0);
	}

	public bool HasAttr(PyObject name) {
	    return (Runtime.PyObject_HasAttr(handle, name.handle) != 0);
	}

	public PyObject GetAttr(string name) {
	    PyString pyName = new PyString(name);
	    IntPtr result = Runtime.PyObject_GetAttr(handle, pyName.handle);
	    if (result == IntPtr.Zero) {
		throw new PythonException();
	    }
	    return new PyObject(result);
	}

	public PyObject GetAttr(string name, PyObject def) {
	    PyString pyName = new PyString(name);
	    IntPtr result = Runtime.PyObject_GetAttr(handle, pyName.handle);
	    if (result == IntPtr.Zero) {
		Runtime.PyErr_Clear();
		return def;
	    }
	    return new PyObject(result);
	}

	public PyObject GetAttr(PyObject name) {
	    IntPtr result = Runtime.PyObject_GetAttr(handle, name.handle);
	    if (result == IntPtr.Zero) {
		throw new PythonException();
	    }
	    return new PyObject(result);
	}

	public PyObject GetAttr(PyObject name, PyObject def) {
	    IntPtr result = Runtime.PyObject_GetAttr(handle, name.handle);
	    if (result == IntPtr.Zero) {
		Runtime.PyErr_Clear();
		return def;
	    }
	    return new PyObject(result);
	}


	public void SetAttr(string name, PyObject value) {
	    int result = Runtime.PyObject_SetAttrString(
				 handle, name, value.handle
				 );
	    if (result < 0) {
		throw new PythonException();
	    }
	}

	public void SetAttr(PyObject name, PyObject value) {
	    int result = Runtime.PyObject_SetAttr(
				 handle, name.handle, value.handle
				 );
	    if (result < 0) {
		throw new PythonException();
	    }
	}

	public void DelAttr(string name) {
	    int result = Runtime.PyObject_SetAttrString(
				 handle, name, IntPtr.Zero
				 );
	    if (result < 0) {
		throw new PythonException();
	    }
	}

	public void DelAttr(PyObject name, PyObject value) {
	    int result = Runtime.PyObject_SetAttr(
				 handle, name.handle, IntPtr.Zero
				 );
	    if (result < 0) {
		throw new PythonException();
	    }
	}



	public virtual PyObject this[string key] {
	    get { return GetItem(key); }
	    set { SetItem(key, value); }
	}

	public virtual PyObject this[PyObject key] {
	    get { return GetItem(key); }
	    set { SetItem(key, value); }
	}

	public virtual PyObject this[int index] {
	    get { return GetItem(index); }
	    set { SetItem(index, value); }
	}

	public virtual PyObject GetItem(PyObject key) {
	    IntPtr result = Runtime.PyObject_GetItem(handle, key.handle);
	    if (result == IntPtr.Zero) {
		throw new PythonException();
	    }
	    return new PyObject(result);
	}

	public virtual PyObject GetItem(string key) {
	    return GetItem(new PyString(key));
	}

	public virtual PyObject GetItem(int index) {
	    PyInteger key = new PyInteger(index);
	    return GetItem((PyObject)key);
	}


	public virtual void SetItem(PyObject key, PyObject value) {
	    int result = Runtime.PyObject_SetItem(
				 handle, key.handle, value.handle
				 );
	    if (result < 0) {
		throw new PythonException();
	    }
	}

	public virtual void SetItem(string key, PyObject value) {
	    SetItem(new PyString(key), value);
	}

	public virtual void SetItem(int index, PyObject value) {
	    // TODO
	    return;
	}


	public virtual void DelItem(PyObject key) {
	    int result = Runtime.PyObject_DelItem(handle, key.handle);
	    if (result < 0) {
		throw new PythonException();
	    }
	}

	public virtual void DelItem(string key) {
	    DelItem(new PyString(key));
	}

	public virtual void DelItem(int index) {
	    // TODO
	    return;
	}


	// TODO: add format-based implementations?

	public PyObject Invoke(PyObject[] args, PyDict kw) {
	    PyTuple t = new PyTuple(args);
	    IntPtr result = Runtime.PyObject_Call(handle, t.handle, kw.handle);
	    if (result == IntPtr.Zero) {
		throw new PythonException();
	    }
	    return new PyObject(result);
	}

	public PyObject Invoke(params PyObject[] args) {
	    PyTuple t = new PyTuple(args);
	    IntPtr result = Runtime.PyObject_Call(handle, t.handle, 
						  IntPtr.Zero
						  );
	    if (result == IntPtr.Zero) {
		throw new PythonException();
	    }
	    return new PyObject(result);
	}

	public PyObject InvokeMethod(string name, PyObject[] args, PyDict kw) {
	    return GetAttr(name).Invoke(args, kw);
	}

	public PyObject InvokeMethod(string name, params PyObject[] args) {
	    return GetAttr(name).Invoke(args);
	}







	public bool IsInstance(PyObject typeOrClass) {
	    int result = Runtime.PyObject_IsInstance(
				 handle, typeOrClass.handle
				 );
	    if (result < 0) {
		Runtime.PyErr_Clear();
		return false;
	    }
	    return (result != 0);
	}

	public bool IsSubclass(PyObject typeOrClass) {
	    int result = Runtime.PyObject_IsSubclass(
				 handle, typeOrClass.handle
				 );
	    if (result < 0) {
		Runtime.PyErr_Clear();
		return false;
	    }
	    return (result != 0);
	}

	public bool IsCallable() {
	    return (Runtime.PyCallable_Check(handle) != 0);
	}

	public bool IsTrue() {
	    return (Runtime.PyObject_IsTrue(handle) != 0);
	}

	public string Repr() {
	    IntPtr strval = Runtime.PyObject_Repr(handle);
	    string result = Runtime.PyString_AsString(strval);
	    //	Runtime.Decref(strval);
	    return result;
	}

	public override string ToString() {
	    IntPtr strval = Runtime.PyObject_Str(handle);
	    string result = Runtime.PyString_AsString(strval);
	    Runtime.Decref(strval);
	    return result;
	}

	public override bool Equals(object o) {
	    if (!(o is PyObject)) {
		return false;
	    }
	    if (handle == ((PyObject) o).handle) {
		return true;
	    }
	    int result = Runtime.PyObject_Compare(
				 handle, ((PyObject) o).handle
				 );
	    if (Exceptions.ErrorOccurred()) {
		throw new PythonException();
	    }
	    return (result == 0);
	}

	public override int GetHashCode() {
	    // fix this!
	    return handle.GetHashCode();
	}

    }


}
