// Copyright (c) 2001, 2002 Zope Corporation and Contributors.
//
// All Rights Reserved.
//
// This software is subject to the provisions of the Zope Public License,
// Version 2.0 (ZPL).  A copy of the ZPL should accompany this distribution.
// THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
// WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
// WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
// FOR A PARTICULAR PURPOSE.

using System;

namespace Python.Runtime {

    /// <summary>
    /// Provides an interface to exceptions thrown by the Python runtime.
    ///
    /// Managed wrappers for the Python APIs will propagate exceptions
    /// thrown in the Python runtime by throwing a PythonException that
    /// wraps and provides access to the Python exception information.
    /// </summary>

    public class PythonException : System.Exception {

	private IntPtr excType = IntPtr.Zero;
	private IntPtr excValue = IntPtr.Zero;
	private IntPtr excTb = IntPtr.Zero;

	public PythonException() : base() {
	    // According to the SDK docs, because IntPtr is blittable the
	    // interop marshaller will pin automatically for byref calls.
	    Runtime.PyErr_Fetch(ref excType, ref excValue, ref excTb);
	}

	/// <summary>
	/// Traceback Property
	/// </summary>
	///
	/// <remarks>
	/// Returns the message associated with the exception.
	/// </remarks>

	string _message;

	public override string Message {
	    get {
		if (_message != null) {
		    return _message;
		}
		IntPtr ob = Runtime.PyObject_Str(excValue);
		_message = Runtime.PyString_AsString(ob);
		return _message;
	    }
	}

	/// <summary>
	/// Type Property
	/// </summary>
	///
	/// <remarks>
	/// Returns the exception type as a Python object.
	/// </remarks>

	public PyObject Type {
	    get {
		return null;
	    }
	}

	/// <summary>
	/// Value Property
	/// </summary>
	///
	/// <remarks>
	/// Returns the exception value as a Python object.
	/// </remarks>

	public PyObject Value {
	    get {
		return null;
	    }
	}

	/// <summary>
	/// Traceback Property
	/// </summary>
	///
	/// <remarks>
	/// Returns the exception traceback as a Python object.
	/// </remarks>

	public PyObject Traceback {
	    get {
		return null;
	    }
	}

	/// <summary>
	/// Dispose Method
	/// </summary>
	///
	/// <remarks>
	/// The Dispose method provides a way to explicitly release the 
	/// Python objects associated with the current Python exception.
	/// The PythonException cannot be used after calling Dispose().
	/// </remarks>

	public void Dispose() {
	    GC.SuppressFinalize(this);
	    Runtime.Decref(excType);
	    Runtime.Decref(excValue);
	    Runtime.Decref(excTb);
	}

	/// <summary>
	/// Finalize Method
	/// </summary>
	///
	/// <remarks>
	/// Ensures that a PythonException releases the Python objects
	/// associated with an exception when it is garbage-collected.
	/// </remarks>

	public void Finalize() {
	    Dispose();
	}

	/// <summary>
	/// Matches Method
	/// </summary>
	///
	/// <remarks>
	/// Returns true if the Python exception type represented by the 
	/// PythonException instance matches the given exception type.
	/// </remarks>

	public static bool Matches(PyObject ob) {
	    return Runtime.PyErr_ExceptionMatches(ob.Handle) != 0;
	}

    } 


}
