// Copyright (c) 2001, 2002 Zope Corporation and Contributors.
//
// All Rights Reserved.
//
// This software is subject to the provisions of the Zope Public License,
// Version 2.0 (ZPL).  A copy of the ZPL should accompany this distribution.
// THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
// WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
// WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
// FOR A PARTICULAR PURPOSE.

using System;
using System.Runtime.InteropServices;
using System.Security;

namespace Python.Runtime {

    [SuppressUnmanagedCodeSecurityAttribute()]

    public class Runtime {

	/// <summary>
	/// Encapsulates the low-level Python C API.
	/// </summary>

	private static IntPtr MainThreadState;

	internal static void Initialize() {

	    Runtime.Py_Initialize();
	    Runtime.PyEval_InitThreads();

	    // Store a reference to the main threadstate.
	    MainThreadState = PyThreadState_Get();


	    IntPtr dict = Runtime.PyImport_GetModuleDict();
	    IntPtr op = Runtime.PyDict_GetItemString(dict, "__builtin__");

	    PyModuleType = Runtime.PyObject_Type(op);
	    PyNone = Runtime.PyObject_GetAttrString(op, "None");
	    PyNoneType = Runtime.PyObject_Type(PyNone);
	    PyTypeType = Runtime.PyObject_Type(PyNoneType);

	    op = Runtime.PyString_FromString("string");
	    PyStringType = Runtime.PyObject_Type(op);
	    Runtime.Decref(op);

	    op = Runtime.PyUnicode_FromString("unicode");
	    PyUnicodeType = Runtime.PyObject_Type(op);
	    Runtime.Decref(op);

	    op = Runtime.PyTuple_New(0);
	    PyTupleType = Runtime.PyObject_Type(op);
	    Runtime.Decref(op);

	    op = Runtime.PyList_New(0);
	    PyListType = Runtime.PyObject_Type(op);
	    Runtime.Decref(op);

	    op = Runtime.PyDict_New();
	    PyDictType = Runtime.PyObject_Type(op);
	    Runtime.Decref(op);

	    op = Runtime.PyInt_FromLong(0);
	    PyIntType = Runtime.PyObject_Type(op);
	    Runtime.Decref(op);

	    op = Runtime.PyLong_FromLong(0);
	    PyLongType = Runtime.PyObject_Type(op);
	    Runtime.Decref(op);

	    op = Runtime.PyFloat_FromDouble(0);
	    PyFloatType = Runtime.PyObject_Type(op);
	    Runtime.Decref(op);

	    IntPtr s = Runtime.PyString_FromString("__temp");
	    IntPtr t = Runtime.PyTuple_New(0);
	    IntPtr d = Runtime.PyDict_New();
	    IntPtr c = Runtime.PyClass_New(t, d, s);

	    PyClassType = Runtime.PyObject_Type(c);
	    IntPtr i = Runtime.PyInstance_New(c, t, d);
	    PyInstanceType = Runtime.PyObject_Type(i);

	    Runtime.Decref(s);
	    Runtime.Decref(t);
	    Runtime.Decref(d);
	    Runtime.Decref(i);
	    Runtime.Decref(c);

	    Error = new IntPtr(-1);

	    PyCLRMetaType = MetaType.Initialize();
	    Exceptions.Initialize();
	    ImportHook.Initialize();


	    // TODO: get rid of this! This works around the gc crash
	    // until I have time to dig into it.

	    IntPtr nm = Runtime.PyString_FromString("gc");
	    IntPtr md = Runtime.PyImport_Import(nm);
	    IntPtr fn = Runtime.PyObject_GetAttrString(md, "disable");
	    IntPtr aa = Runtime.PyTuple_New(0);
	    IntPtr rr = Runtime.PyObject_Call(fn, aa, IntPtr.Zero);
	    Runtime.Decref(rr);
	    Runtime.Decref(aa);
	    Runtime.Decref(fn);
	    Runtime.Decref(nm);
	    Runtime.Decref(md);
	}

	internal unsafe static void Incref(IntPtr op) {
	    if (op == IntPtr.Zero) 
		return;
	    (*(int *)op)++;
	}

	private static void LogDecref(IntPtr op) {
	    IntPtr tp = Runtime.PyObject_Type(op);
	    IntPtr fp = Marshal.ReadIntPtr(tp, (6 * IntPtr.Size));
	    if (fp == IntPtr.Zero) {
		return;
	    }

	    IntPtr refs = Marshal.ReadIntPtr(op);

	    IntPtr p1 = Runtime.PyObject_Repr(tp);
	    string s1 = Runtime.PyString_AsString(p1);
	    
	    IntPtr p2 = Runtime.PyObject_Repr(op);
	    string s2 = Runtime.PyString_AsString(p2);

	    Console.WriteLine("decref: {0} {1} {2} {3} {4}", 
			      s1, refs, s2, fp, op
			      );
	}

	internal unsafe static void Decref(IntPtr op) {

	    if (op == IntPtr.Zero) {
		return;
	    }

	    if (--(*(int *)op) == 0) {

		// fp = op -> ob_type -> tp_dealloc
		IntPtr tp = Runtime.PyObject_Type(op);
		IntPtr fp = Marshal.ReadIntPtr(tp, (6 * IntPtr.Size));
		if (fp == IntPtr.Zero) {
		    return;
		}
		//LogDecref(op);
		try {
		    // Fix this - figure out native call / malloc bug.
		    // NativeCall.Impl.Void_Call_1(fp, op);
		} catch {
		    IntPtr ss = Runtime.PyObject_Str(op);
		    string ds = Runtime.PyString_AsString(ss);
		    Console.WriteLine("native call failed!");
		}

		return;
	    }

	}

	internal unsafe static int GetRefCount(IntPtr op) {
	    if (op == IntPtr.Zero) 
		return 0;
	    return *(int *)op;
	}

	// Utility to return the name of an object's type as a string.

	internal static string PyObject_GetTypeName(IntPtr op) {
	    IntPtr pyType = Marshal.ReadIntPtr(op, IntPtr.Size);
	    IntPtr ppName = Marshal.ReadIntPtr(pyType, (3 * IntPtr.Size));
	    return Marshal.PtrToStringAnsi(ppName);
	}

	internal static IntPtr PyModuleType;
	internal static IntPtr PyClassType;
	internal static IntPtr PyInstanceType;
	internal static IntPtr PyCLRMetaType;

	internal static IntPtr PyUnicodeType;
	internal static IntPtr PyStringType;
	internal static IntPtr PyTupleType;
	internal static IntPtr PyListType;	
	internal static IntPtr PyDictType;	
	internal static IntPtr PyIntType;
	internal static IntPtr PyLongType;
	internal static IntPtr PyFloatType;
	internal static IntPtr PyNoneType;
	internal static IntPtr PyTypeType;

	internal static IntPtr PyNone;
	internal static IntPtr Error;


	[DllImport("python22", CallingConvention=CallingConvention.Cdecl,
		    ExactSpelling=true, CharSet=CharSet.Ansi)]
	internal unsafe static extern void
	Py_Initialize();

	[DllImport("python22", CallingConvention=CallingConvention.Cdecl,
		    ExactSpelling=true, CharSet=CharSet.Ansi)]
	internal unsafe static extern void
	Py_Finalize();

	[DllImport("python22", CallingConvention=CallingConvention.Cdecl,
		    ExactSpelling=true, CharSet=CharSet.Ansi)]
	internal unsafe static extern IntPtr
	Py_NewInterpreter();

	[DllImport("python22", CallingConvention=CallingConvention.Cdecl,
		    ExactSpelling=true, CharSet=CharSet.Ansi)]
        internal unsafe static extern void
	Py_EndInterpreter(IntPtr threadState);

	[DllImport("python22", CallingConvention=CallingConvention.Cdecl,
		    ExactSpelling=true, CharSet=CharSet.Ansi)]
	internal unsafe static extern IntPtr
	PyThreadState_Get();

	[DllImport("python22", CallingConvention=CallingConvention.Cdecl,
		    ExactSpelling=true, CharSet=CharSet.Ansi)]
	public unsafe static extern int
	Py_Main(int argc, string[] argv);


	// Thread state operations

	[DllImport("python22", CallingConvention=CallingConvention.Cdecl,
		    ExactSpelling=true, CharSet=CharSet.Ansi)]
	internal unsafe static extern void
	PyEval_InitThreads();

	[DllImport("python22", CallingConvention=CallingConvention.Cdecl,
		    ExactSpelling=true, CharSet=CharSet.Ansi)]
	internal unsafe static extern void
	PyEval_AcquireLock();

	[DllImport("python22", CallingConvention=CallingConvention.Cdecl,
		    ExactSpelling=true, CharSet=CharSet.Ansi)]
	internal unsafe static extern void
	PyEval_ReleaseLock();

	[DllImport("python22", CallingConvention=CallingConvention.Cdecl,
		    ExactSpelling=true, CharSet=CharSet.Ansi)]
	internal unsafe static extern void
	PyEval_AcquireThread(IntPtr tstate);

	[DllImport("python22", CallingConvention=CallingConvention.Cdecl,
		    ExactSpelling=true, CharSet=CharSet.Ansi)]
	internal unsafe static extern void
	PyEval_ReleaseThread(IntPtr tstate);

	[DllImport("python22", CallingConvention=CallingConvention.Cdecl,
		    ExactSpelling=true, CharSet=CharSet.Ansi)]
	internal unsafe static extern IntPtr
	PyEval_SaveThread();

	[DllImport("python22", CallingConvention=CallingConvention.Cdecl,
		    ExactSpelling=true, CharSet=CharSet.Ansi)]
	internal unsafe static extern void
	PyEval_RestoreThread(IntPtr tstate);


	// Miscellaneous high-level runtime APIs

	[DllImport("python22", CallingConvention=CallingConvention.Cdecl,
		    ExactSpelling=true, CharSet=CharSet.Ansi)]
        internal unsafe static extern string
        Py_GetProgramName();

	[DllImport("python22", CallingConvention=CallingConvention.Cdecl,
		    ExactSpelling=true, CharSet=CharSet.Ansi)]
        internal unsafe static extern void
        Py_SetProgramName(string name);

	[DllImport("python22", CallingConvention=CallingConvention.Cdecl,
		    ExactSpelling=true, CharSet=CharSet.Ansi)]
        internal unsafe static extern string
        Py_GetPythonHome();

	[DllImport("python22", CallingConvention=CallingConvention.Cdecl,
		    ExactSpelling=true, CharSet=CharSet.Ansi)]
        internal unsafe static extern void
        Py_SetPythonHome(string home);

	[DllImport("python22", CallingConvention=CallingConvention.Cdecl,
		    ExactSpelling=true, CharSet=CharSet.Ansi)]
        internal unsafe static extern string
        Py_GetVersion();

	[DllImport("python22", CallingConvention=CallingConvention.Cdecl,
		    ExactSpelling=true, CharSet=CharSet.Ansi)]
        internal unsafe static extern string
        Py_GetPlatform();

	[DllImport("python22", CallingConvention=CallingConvention.Cdecl,
		    ExactSpelling=true, CharSet=CharSet.Ansi)]
        internal unsafe static extern string
        Py_GetCopyright();

	[DllImport("python22", CallingConvention=CallingConvention.Cdecl,
		    ExactSpelling=true, CharSet=CharSet.Ansi)]
        internal unsafe static extern string
        Py_GetCompiler();

	[DllImport("python22", CallingConvention=CallingConvention.Cdecl,
		    ExactSpelling=true, CharSet=CharSet.Ansi)]
        internal unsafe static extern string
        Py_GetBuildInfo();

	[DllImport("python22", CallingConvention=CallingConvention.Cdecl,
		    ExactSpelling=true, CharSet=CharSet.Ansi)]
        internal unsafe static extern int
        PyRun_SimpleString(string code);


	// Python abstract object API

	[DllImport("python22", CallingConvention=CallingConvention.Cdecl,
		    ExactSpelling=true, CharSet=CharSet.Ansi)]
        internal unsafe static extern IntPtr
        PyObject_Type(IntPtr pointer);

	[DllImport("python22", CallingConvention=CallingConvention.Cdecl,
		    ExactSpelling=true, CharSet=CharSet.Ansi)]
        internal unsafe static extern int
        PyObject_HasAttrString(IntPtr pointer, string name);

	[DllImport("python22", CallingConvention=CallingConvention.Cdecl,
		    ExactSpelling=true, CharSet=CharSet.Ansi)]
        internal unsafe static extern IntPtr
        PyObject_GetAttrString(IntPtr pointer, string name);

	[DllImport("python22", CallingConvention=CallingConvention.Cdecl,
		    ExactSpelling=true, CharSet=CharSet.Ansi)]
        internal unsafe static extern int
        PyObject_SetAttrString(IntPtr pointer, string name, IntPtr value);

	[DllImport("python22", CallingConvention=CallingConvention.Cdecl,
		    ExactSpelling=true, CharSet=CharSet.Ansi)]
        internal unsafe static extern int
        PyObject_HasAttr(IntPtr pointer, IntPtr name);

	[DllImport("python22", CallingConvention=CallingConvention.Cdecl,
		    ExactSpelling=true, CharSet=CharSet.Ansi)]
        internal unsafe static extern IntPtr
        PyObject_GetAttr(IntPtr pointer, IntPtr name);

	[DllImport("python22", CallingConvention=CallingConvention.Cdecl,
		    ExactSpelling=true, CharSet=CharSet.Ansi)]
        internal unsafe static extern int
        PyObject_SetAttr(IntPtr pointer, IntPtr name, IntPtr value);

	[DllImport("python22", CallingConvention=CallingConvention.Cdecl,
		    ExactSpelling=true, CharSet=CharSet.Ansi)]
        internal unsafe static extern IntPtr
        PyObject_GetItem(IntPtr pointer, IntPtr key);

	[DllImport("python22", CallingConvention=CallingConvention.Cdecl,
		    ExactSpelling=true, CharSet=CharSet.Ansi)]
        internal unsafe static extern int
        PyObject_SetItem(IntPtr pointer, IntPtr key, IntPtr value);

	[DllImport("python22", CallingConvention=CallingConvention.Cdecl,
		    ExactSpelling=true, CharSet=CharSet.Ansi)]
        internal unsafe static extern int
        PyObject_DelItem(IntPtr pointer, IntPtr key);

	[DllImport("python22", CallingConvention=CallingConvention.Cdecl,
		    ExactSpelling=true, CharSet=CharSet.Ansi)]
        internal unsafe static extern IntPtr
        PyObject_Call(IntPtr pointer, IntPtr args, IntPtr kw);

	[DllImport("python22", CallingConvention=CallingConvention.Cdecl,
		    ExactSpelling=true, CharSet=CharSet.Ansi)]
	internal unsafe static extern IntPtr
	PyObject_CallObject(IntPtr pointer, IntPtr args);

	

	[DllImport("python22", CallingConvention=CallingConvention.Cdecl,
		    ExactSpelling=true, CharSet=CharSet.Ansi)]
	internal unsafe static extern int
	PyObject_Compare(IntPtr value1, IntPtr value2);

	[DllImport("python22", CallingConvention=CallingConvention.Cdecl,
		    ExactSpelling=true, CharSet=CharSet.Ansi)]
	internal unsafe static extern int
	PyObject_IsInstance(IntPtr ob, IntPtr type);

	[DllImport("python22", CallingConvention=CallingConvention.Cdecl,
		    ExactSpelling=true, CharSet=CharSet.Ansi)]
	internal unsafe static extern int
	PyObject_IsSubclass(IntPtr ob, IntPtr type);

	[DllImport("python22", CallingConvention=CallingConvention.Cdecl,
		    ExactSpelling=true, CharSet=CharSet.Ansi)]
	internal unsafe static extern int
	PyCallable_Check(IntPtr pointer);

	[DllImport("python22", CallingConvention=CallingConvention.Cdecl,
		    ExactSpelling=true, CharSet=CharSet.Ansi)]
	internal unsafe static extern int
	PyObject_IsTrue(IntPtr pointer);

	[DllImport("python22", CallingConvention=CallingConvention.Cdecl,
		    ExactSpelling=true, CharSet=CharSet.Ansi)]
	internal unsafe static extern int
	PyObject_Size(IntPtr pointer);

	[DllImport("python22", CallingConvention=CallingConvention.Cdecl,
		    ExactSpelling=true, CharSet=CharSet.Ansi)]
	internal unsafe static extern IntPtr
	PyObject_Hash(IntPtr op);

	[DllImport("python22", CallingConvention=CallingConvention.Cdecl,
		    ExactSpelling=true, CharSet=CharSet.Ansi)]
	internal unsafe static extern IntPtr
	PyObject_Repr(IntPtr pointer);

	[DllImport("python22", CallingConvention=CallingConvention.Cdecl,
		    ExactSpelling=true, CharSet=CharSet.Ansi)]
	internal unsafe static extern IntPtr
	PyObject_Str(IntPtr pointer);


	// Python numeric methods

	[DllImport("python22", CallingConvention=CallingConvention.Cdecl,
		    ExactSpelling=true, CharSet=CharSet.Ansi)]
	internal unsafe static extern IntPtr
	PyNumber_Int(IntPtr ob);

	[DllImport("python22", CallingConvention=CallingConvention.Cdecl,
		    ExactSpelling=true, CharSet=CharSet.Ansi)]
	internal unsafe static extern IntPtr
	PyNumber_Long(IntPtr ob);

	[DllImport("python22", CallingConvention=CallingConvention.Cdecl,
		    ExactSpelling=true, CharSet=CharSet.Ansi)]
	internal unsafe static extern IntPtr
	PyNumber_Float(IntPtr ob);



	[DllImport("python22", CallingConvention=CallingConvention.Cdecl,
		    ExactSpelling=true, CharSet=CharSet.Ansi)]
	internal unsafe static extern IntPtr
	PyInt_FromLong(int value);

	[DllImport("python22", CallingConvention=CallingConvention.Cdecl,
		    ExactSpelling=true, CharSet=CharSet.Ansi)]
	internal unsafe static extern int
	PyInt_AsLong(IntPtr value);

	[DllImport("python22", CallingConvention=CallingConvention.Cdecl,
		    ExactSpelling=true, CharSet=CharSet.Ansi)]
	internal unsafe static extern IntPtr
	PyInt_FromString(string value, IntPtr end, int radix);

	[DllImport("python22", CallingConvention=CallingConvention.Cdecl,
		    ExactSpelling=true, CharSet=CharSet.Ansi)]
	internal unsafe static extern int
	PyInt_GetMax();


	[DllImport("python22", CallingConvention=CallingConvention.Cdecl,
		    ExactSpelling=true, CharSet=CharSet.Ansi)]
	internal unsafe static extern IntPtr
	PyLong_FromLong(long value);

	[DllImport("python22", CallingConvention=CallingConvention.Cdecl,
		    ExactSpelling=true, CharSet=CharSet.Ansi)]
	internal unsafe static extern IntPtr
	PyLong_FromUnsignedLong(uint value);

	[DllImport("python22", CallingConvention=CallingConvention.Cdecl,
		    ExactSpelling=true, CharSet=CharSet.Ansi)]
	internal unsafe static extern IntPtr
	PyLong_FromDouble(double value);

	[DllImport("python22", CallingConvention=CallingConvention.Cdecl,
		    ExactSpelling=true, CharSet=CharSet.Ansi)]
	internal unsafe static extern IntPtr
	PyLong_FromLongLong(long value);

	[DllImport("python22", CallingConvention=CallingConvention.Cdecl,
		    ExactSpelling=true, CharSet=CharSet.Ansi)]
	internal unsafe static extern IntPtr
	PyLong_FromUnsignedLongLong(ulong value);


	[DllImport("python22", CallingConvention=CallingConvention.Cdecl,
		    ExactSpelling=true, CharSet=CharSet.Ansi)]
	internal unsafe static extern IntPtr
	PyLong_FromString(string value, IntPtr end, int radix);

	[DllImport("python22", CallingConvention=CallingConvention.Cdecl,
		    ExactSpelling=true, CharSet=CharSet.Ansi)]
	internal unsafe static extern int
	PyLong_AsLong(IntPtr value);

	[DllImport("python22", CallingConvention=CallingConvention.Cdecl,
		    ExactSpelling=true, CharSet=CharSet.Ansi)]
	internal unsafe static extern uint
	PyLong_AsUnsignedLong(IntPtr value);

	[DllImport("python22", CallingConvention=CallingConvention.Cdecl,
		    ExactSpelling=true, CharSet=CharSet.Ansi)]
	internal unsafe static extern long
	PyLong_AsLongLong(IntPtr value);

	[DllImport("python22", CallingConvention=CallingConvention.Cdecl,
		    ExactSpelling=true, CharSet=CharSet.Ansi)]
	internal unsafe static extern ulong
	PyLong_AsUnsignedLongLong(IntPtr value);



	[DllImport("python22", CallingConvention=CallingConvention.Cdecl,
		    ExactSpelling=true, CharSet=CharSet.Ansi)]
	internal unsafe static extern IntPtr
	PyFloat_FromDouble(double value);

	[DllImport("python22", CallingConvention=CallingConvention.Cdecl,
		    ExactSpelling=true, CharSet=CharSet.Ansi)]
	internal unsafe static extern IntPtr
	PyFloat_FromString(IntPtr value, IntPtr junk);

	[DllImport("python22", CallingConvention=CallingConvention.Cdecl,
		    ExactSpelling=true, CharSet=CharSet.Ansi)]
	internal unsafe static extern double
	PyFloat_AsDouble(IntPtr ob);







	// Python abstract sequence API

	[DllImport("python22", CallingConvention=CallingConvention.Cdecl,
		    ExactSpelling=true, CharSet=CharSet.Ansi)]
	internal unsafe static extern int
	PySequence_Check(IntPtr pointer);

	[DllImport("python22", CallingConvention=CallingConvention.Cdecl,
		    ExactSpelling=true, CharSet=CharSet.Ansi)]
	internal unsafe static extern IntPtr
	PySequence_GetItem(IntPtr pointer, int index);

	[DllImport("python22", CallingConvention=CallingConvention.Cdecl,
		    ExactSpelling=true, CharSet=CharSet.Ansi)]
	internal unsafe static extern int
	PySequence_SetItem(IntPtr pointer, int index, IntPtr value);

	[DllImport("python22", CallingConvention=CallingConvention.Cdecl,
		    ExactSpelling=true, CharSet=CharSet.Ansi)]
	internal unsafe static extern int
	PySequence_DelItem(IntPtr pointer, int index);


	[DllImport("python22", CallingConvention=CallingConvention.Cdecl,
		    ExactSpelling=true, CharSet=CharSet.Ansi)]
	internal unsafe static extern int
	PySequence_Size(IntPtr pointer);

	[DllImport("python22", CallingConvention=CallingConvention.Cdecl,
		    ExactSpelling=true, CharSet=CharSet.Ansi)]
	internal unsafe static extern int
	PySequence_Contains(IntPtr pointer, IntPtr item);

	[DllImport("python22", CallingConvention=CallingConvention.Cdecl,
		    ExactSpelling=true, CharSet=CharSet.Ansi)]
	internal unsafe static extern IntPtr
	PySequence_Concat(IntPtr pointer, IntPtr other);

	[DllImport("python22", CallingConvention=CallingConvention.Cdecl,
		    ExactSpelling=true, CharSet=CharSet.Ansi)]
	internal unsafe static extern IntPtr
	PySequence_Repeat(IntPtr pointer, int count);

	[DllImport("python22", CallingConvention=CallingConvention.Cdecl,
		    ExactSpelling=true, CharSet=CharSet.Ansi)]
	internal unsafe static extern int
	PySequence_Index(IntPtr pointer, IntPtr item);

	[DllImport("python22", CallingConvention=CallingConvention.Cdecl,
		    ExactSpelling=true, CharSet=CharSet.Ansi)]
	internal unsafe static extern int
	PySequence_Count(IntPtr pointer, IntPtr value);

	[DllImport("python22", CallingConvention=CallingConvention.Cdecl,
		    ExactSpelling=true, CharSet=CharSet.Ansi)]
	internal unsafe static extern IntPtr
	PySequence_Tuple(IntPtr pointer);

	[DllImport("python22", CallingConvention=CallingConvention.Cdecl,
		    ExactSpelling=true, CharSet=CharSet.Ansi)]
	internal unsafe static extern IntPtr
	PySequence_List(IntPtr pointer);



	// Python string methods

	internal static bool IsStringType(IntPtr op) {
	    IntPtr t = PyObject_Type(op);
	    return (t == PyStringType) || (t == PyUnicodeType);
	}

	internal static bool PyString_Check(IntPtr ob) {
	    return PyObject_Type(ob) == Runtime.PyStringType;
	}

	[DllImport("python22", CallingConvention=CallingConvention.Cdecl,
		    ExactSpelling=true, CharSet=CharSet.Ansi)]
	internal unsafe static extern IntPtr
	PyString_FromString(string value);

	[DllImport("python22", CallingConvention=CallingConvention.Cdecl,
		    ExactSpelling=true, CharSet=CharSet.Ansi)]
	internal unsafe static extern IntPtr
	PyString_FromStringAndSize(string value, int size);

	[DllImport("python22", CallingConvention=CallingConvention.Cdecl,
		    ExactSpelling=true, CharSet=CharSet.Ansi)]
	internal unsafe static extern string
	PyString_AsString(IntPtr pointer);

	[DllImport("python22", CallingConvention=CallingConvention.Cdecl,
		   EntryPoint="PyString_AsString",
		   ExactSpelling=true, CharSet=CharSet.Ansi)]
	internal unsafe static extern IntPtr 
	PyString_AS_STRING(IntPtr op);

	[DllImport("python22", CallingConvention=CallingConvention.Cdecl,
		    ExactSpelling=true, CharSet=CharSet.Ansi)]
	internal unsafe static extern int
	PyString_Size(IntPtr pointer);


	// Python unicode methods. Note that these support UCS-2 only!

	internal static bool PyUnicode_Check(IntPtr ob) {
	    return PyObject_Type(ob) == Runtime.PyUnicodeType;
	}

	[DllImport("python22", CallingConvention=CallingConvention.Cdecl,
		   EntryPoint="PyUnicodeUCS2_FromObject",
		    ExactSpelling=true, CharSet=CharSet.Unicode)]
	internal unsafe static extern IntPtr
	PyUnicode_FromObject(IntPtr ob);

	[DllImport("python22", CallingConvention=CallingConvention.Cdecl,
		   EntryPoint="PyUnicodeUCS2_FromEncodedObject",
		    ExactSpelling=true, CharSet=CharSet.Unicode)]
	internal unsafe static extern IntPtr
	PyUnicode_FromEncodedObject(IntPtr ob, IntPtr enc, IntPtr err);

	[DllImport("python22", CallingConvention=CallingConvention.Cdecl,
		   EntryPoint="PyUnicodeUCS2_FromUnicode",
		   ExactSpelling=true, CharSet=CharSet.Unicode)]
	internal unsafe static extern IntPtr
	PyUnicode_FromUnicode(string s, int size);

	internal static IntPtr PyUnicode_FromString(string s) {
	    return PyUnicode_FromUnicode(s, (s.Length));
	}

	[DllImport("python22", CallingConvention=CallingConvention.Cdecl,
		   EntryPoint="PyUnicodeUCS2_GetSize",
		   ExactSpelling=true, CharSet=CharSet.Ansi)]
	internal unsafe static extern int
	PyUnicode_GetSize(IntPtr ob);

	[DllImport("python22", CallingConvention=CallingConvention.Cdecl,
		   EntryPoint="PyUnicodeUCS2_AsUnicode",
		   ExactSpelling=true, CharSet=CharSet.Unicode)]
	internal unsafe static extern string
	PyUnicode_AsUnicode(IntPtr ob);

	[DllImport("python22", CallingConvention=CallingConvention.Cdecl,
		   EntryPoint="PyUnicodeUCS2_AsUnicode",
		   ExactSpelling=true, CharSet=CharSet.Unicode)]
	internal unsafe static extern IntPtr
	PyUnicode_AS_UNICODE(IntPtr op);

	[DllImport("python22", CallingConvention=CallingConvention.Cdecl,
		   ExactSpelling=true, CharSet=CharSet.Unicode)]
	internal unsafe static extern IntPtr 
	PyUnicode_FromOrdinal(int c);

	internal static string GetManagedString(IntPtr op) {
	    IntPtr type = PyObject_Type(op);
	    if (type == Runtime.PyStringType) {
		return Runtime.PyString_AsString(op);
	    }
	    if (type == Runtime.PyUnicodeType) {
		return Runtime.PyUnicode_AsUnicode(op);
	    }
	    return null;
	}


	// Python dict methods

	internal static bool PyDict_Check(IntPtr ob) {
	    return PyObject_Type(ob) == Runtime.PyDictType;
	}

	[DllImport("python22", CallingConvention=CallingConvention.Cdecl,
		    ExactSpelling=true, CharSet=CharSet.Ansi)]
	internal unsafe static extern IntPtr
	PyDict_New();

	[DllImport("python22", CallingConvention=CallingConvention.Cdecl,
		    ExactSpelling=true, CharSet=CharSet.Ansi)]
	internal unsafe static extern IntPtr
	PyDictProxy_New(IntPtr dict);

	[DllImport("python22", CallingConvention=CallingConvention.Cdecl,
		    ExactSpelling=true, CharSet=CharSet.Ansi)]
	internal unsafe static extern IntPtr
	PyDict_GetItem(IntPtr pointer, IntPtr key);

	[DllImport("python22", CallingConvention=CallingConvention.Cdecl,
		    ExactSpelling=true, CharSet=CharSet.Ansi)]
	internal unsafe static extern IntPtr
	PyDict_GetItemString(IntPtr pointer, string key);

	[DllImport("python22", CallingConvention=CallingConvention.Cdecl,
		    ExactSpelling=true, CharSet=CharSet.Ansi)]
	internal unsafe static extern int
	PyDict_SetItem(IntPtr pointer, IntPtr key, IntPtr value);

	[DllImport("python22", CallingConvention=CallingConvention.Cdecl,
		    ExactSpelling=true, CharSet=CharSet.Ansi)]
	internal unsafe static extern int
	PyDict_SetItemString(IntPtr pointer, string key, IntPtr value);

	[DllImport("python22", CallingConvention=CallingConvention.Cdecl,
		    ExactSpelling=true, CharSet=CharSet.Ansi)]
	internal unsafe static extern int
	PyDict_DelItem(IntPtr pointer, IntPtr key);

	[DllImport("python22", CallingConvention=CallingConvention.Cdecl,
		    ExactSpelling=true, CharSet=CharSet.Ansi)]
	internal unsafe static extern int
	PyMapping_HasKey(IntPtr pointer, IntPtr key);

	[DllImport("python22", CallingConvention=CallingConvention.Cdecl,
		    ExactSpelling=true, CharSet=CharSet.Ansi)]
	internal unsafe static extern IntPtr
	PyDict_Keys(IntPtr pointer);

	[DllImport("python22", CallingConvention=CallingConvention.Cdecl,
		    ExactSpelling=true, CharSet=CharSet.Ansi)]
	internal unsafe static extern IntPtr
	PyDict_Values(IntPtr pointer);

	[DllImport("python22", CallingConvention=CallingConvention.Cdecl,
		    ExactSpelling=true, CharSet=CharSet.Ansi)]
	internal unsafe static extern IntPtr
	PyDict_Items(IntPtr pointer);

	[DllImport("python22", CallingConvention=CallingConvention.Cdecl,
		    ExactSpelling=true, CharSet=CharSet.Ansi)]
	internal unsafe static extern IntPtr
	PyDict_Copy(IntPtr pointer);

	[DllImport("python22", CallingConvention=CallingConvention.Cdecl,
		    ExactSpelling=true, CharSet=CharSet.Ansi)]
	internal unsafe static extern int
	PyDict_Update(IntPtr pointer, IntPtr other);

	[DllImport("python22", CallingConvention=CallingConvention.Cdecl,
		    ExactSpelling=true, CharSet=CharSet.Ansi)]
	internal unsafe static extern void
	PyDict_Clear(IntPtr pointer);

	[DllImport("python22", CallingConvention=CallingConvention.Cdecl,
		    ExactSpelling=true, CharSet=CharSet.Ansi)]
	internal unsafe static extern int
	PyDict_Size(IntPtr pointer);


	// Python list methods

	internal static bool PyList_Check(IntPtr ob) {
	    return PyObject_Type(ob) == Runtime.PyListType;
	}

	[DllImport("python22", CallingConvention=CallingConvention.Cdecl,
		    ExactSpelling=true, CharSet=CharSet.Ansi)]
	internal unsafe static extern IntPtr
	PyList_New(int size);

	[DllImport("python22", CallingConvention=CallingConvention.Cdecl,
		    ExactSpelling=true, CharSet=CharSet.Ansi)]
	internal unsafe static extern IntPtr
	PyList_AsTuple(IntPtr pointer);

	[DllImport("python22", CallingConvention=CallingConvention.Cdecl,
		    ExactSpelling=true, CharSet=CharSet.Ansi)]
	internal unsafe static extern IntPtr
	PyList_GetItem(IntPtr pointer, int index);

	[DllImport("python22", CallingConvention=CallingConvention.Cdecl,
		    ExactSpelling=true, CharSet=CharSet.Ansi)]
	internal unsafe static extern int
	PyList_SetItem(IntPtr pointer, int index, IntPtr value);

	[DllImport("python22", CallingConvention=CallingConvention.Cdecl,
		    ExactSpelling=true, CharSet=CharSet.Ansi)]
	internal unsafe static extern int
	PyList_Insert(IntPtr pointer, int index, IntPtr value);

	[DllImport("python22", CallingConvention=CallingConvention.Cdecl,
		    ExactSpelling=true, CharSet=CharSet.Ansi)]
	internal unsafe static extern int
	PyList_Append(IntPtr pointer, IntPtr value);

	[DllImport("python22", CallingConvention=CallingConvention.Cdecl,
		    ExactSpelling=true, CharSet=CharSet.Ansi)]
	internal unsafe static extern int
	PyList_Reverse(IntPtr pointer);

	[DllImport("python22", CallingConvention=CallingConvention.Cdecl,
		    ExactSpelling=true, CharSet=CharSet.Ansi)]
	internal unsafe static extern int
	PyList_Sort(IntPtr pointer);

	[DllImport("python22", CallingConvention=CallingConvention.Cdecl,
		    ExactSpelling=true, CharSet=CharSet.Ansi)]
	internal unsafe static extern IntPtr
	PyList_GetSlice(IntPtr pointer, int start, int end);

	[DllImport("python22", CallingConvention=CallingConvention.Cdecl,
		    ExactSpelling=true, CharSet=CharSet.Ansi)]
	internal unsafe static extern int
	PyList_SetSlice(IntPtr pointer, int start, int end, IntPtr value);

	[DllImport("python22", CallingConvention=CallingConvention.Cdecl,
		    ExactSpelling=true, CharSet=CharSet.Ansi)]
	internal unsafe static extern int
	PyList_Size(IntPtr pointer);


	// Python tuple methods

	[DllImport("python22", CallingConvention=CallingConvention.Cdecl,
		    ExactSpelling=true, CharSet=CharSet.Ansi)]
	internal unsafe static extern IntPtr
	PyTuple_New(int size);

	[DllImport("python22", CallingConvention=CallingConvention.Cdecl,
		    ExactSpelling=true, CharSet=CharSet.Ansi)]
	internal unsafe static extern IntPtr
	PyTuple_GetItem(IntPtr pointer, int index);

	[DllImport("python22", CallingConvention=CallingConvention.Cdecl,
		    ExactSpelling=true, CharSet=CharSet.Ansi)]
	internal unsafe static extern int
	PyTuple_SetItem(IntPtr pointer, int index, IntPtr value);

	[DllImport("python22", CallingConvention=CallingConvention.Cdecl,
		    ExactSpelling=true, CharSet=CharSet.Ansi)]
	internal unsafe static extern IntPtr
	PyTuple_GetSlice(IntPtr pointer, int start, int end);

	[DllImport("python22", CallingConvention=CallingConvention.Cdecl,
		    ExactSpelling=true, CharSet=CharSet.Ansi)]
	internal unsafe static extern int
	PyTuple_Size(IntPtr pointer);


	// Python module methods

	[DllImport("python22", CallingConvention=CallingConvention.Cdecl,
		    ExactSpelling=true, CharSet=CharSet.Ansi)]
	internal unsafe static extern string
	PyModule_GetName(IntPtr module);

	[DllImport("python22", CallingConvention=CallingConvention.Cdecl,
		    ExactSpelling=true, CharSet=CharSet.Ansi)]
	internal unsafe static extern IntPtr
	PyModule_GetDict(IntPtr module);

	[DllImport("python22", CallingConvention=CallingConvention.Cdecl,
		    ExactSpelling=true, CharSet=CharSet.Ansi)]
	internal unsafe static extern string
	PyModule_GetFilename(IntPtr module);




	[DllImport("python22", CallingConvention=CallingConvention.Cdecl,
		    ExactSpelling=true, CharSet=CharSet.Ansi)]
	internal unsafe static extern IntPtr
	PyImport_Import(IntPtr name);

	[DllImport("python22", CallingConvention=CallingConvention.Cdecl,
		    ExactSpelling=true, CharSet=CharSet.Ansi)]
	internal unsafe static extern IntPtr
	PyImport_ImportModule(string name);

	[DllImport("python22", CallingConvention=CallingConvention.Cdecl,
		    ExactSpelling=true, CharSet=CharSet.Ansi)]
	internal unsafe static extern IntPtr
	PyImport_ReloadModule(IntPtr module);

	[DllImport("python22", CallingConvention=CallingConvention.Cdecl,
		    ExactSpelling=true, CharSet=CharSet.Ansi)]
	internal unsafe static extern IntPtr
	PyImport_AddModule(string name);

	[DllImport("python22", CallingConvention=CallingConvention.Cdecl,
		    ExactSpelling=true, CharSet=CharSet.Ansi)]
	internal unsafe static extern IntPtr
	PyImport_GetModuleDict();


	[DllImport("python22", CallingConvention=CallingConvention.Cdecl,
		    ExactSpelling=true, CharSet=CharSet.Ansi)]
	internal unsafe static extern void
	PySys_SetArgv(int argc, IntPtr argv);

	[DllImport("python22", CallingConvention=CallingConvention.Cdecl,
		    ExactSpelling=true, CharSet=CharSet.Ansi)]
	internal unsafe static extern IntPtr
	PySys_GetObject(string name);

	[DllImport("python22", CallingConvention=CallingConvention.Cdecl,
		    ExactSpelling=true, CharSet=CharSet.Ansi)]
	internal unsafe static extern int
	PySys_SetObject(string name, IntPtr ob);




	// Type related methods

	[DllImport("python22", CallingConvention=CallingConvention.Cdecl,
		    ExactSpelling=true, CharSet=CharSet.Ansi)]
	internal unsafe static extern bool 
	PyType_IsSubtype(IntPtr t1, IntPtr t2);

        internal static bool PyObject_TypeCheck(IntPtr ob, IntPtr tp) {
	  IntPtr t = PyObject_Type(ob);
	  return (t == tp) || PyType_IsSubtype(t, tp);
	}

	[DllImport("python22", CallingConvention=CallingConvention.Cdecl,
		    ExactSpelling=true, CharSet=CharSet.Ansi)]
	internal unsafe static extern IntPtr
	PyType_GenericNew(IntPtr type, IntPtr args, IntPtr kw);

	[DllImport("python22", CallingConvention=CallingConvention.Cdecl,
		    ExactSpelling=true, CharSet=CharSet.Ansi)]
	internal unsafe static extern IntPtr
	PyType_GenericAlloc(IntPtr type, int n);

	[DllImport("python22", CallingConvention=CallingConvention.Cdecl,
		    ExactSpelling=true, CharSet=CharSet.Ansi)]
	internal unsafe static extern int
	PyType_Ready(IntPtr type);

	[DllImport("python22", CallingConvention=CallingConvention.Cdecl,
		    ExactSpelling=true, CharSet=CharSet.Ansi)]
	internal unsafe static extern IntPtr
	_PyType_Lookup(IntPtr type, IntPtr name);

	[DllImport("python22", CallingConvention=CallingConvention.Cdecl,
		    ExactSpelling=true, CharSet=CharSet.Ansi)]
	internal unsafe static extern IntPtr
	PyObject_GenericGetAttr(IntPtr obj, IntPtr name);

	[DllImport("python22", CallingConvention=CallingConvention.Cdecl,
		    ExactSpelling=true, CharSet=CharSet.Ansi)]
	internal unsafe static extern int
	PyObject_GenericSetAttr(IntPtr obj, IntPtr name, IntPtr value);

	[DllImport("python22", CallingConvention=CallingConvention.Cdecl,
		    ExactSpelling=true, CharSet=CharSet.Ansi)]
	internal unsafe static extern IntPtr
	_PyObject_GetDictPtr(IntPtr obj);


	[DllImport("python22", CallingConvention=CallingConvention.Cdecl,
		    ExactSpelling=true, CharSet=CharSet.Ansi)]
	internal unsafe static extern IntPtr
	PyCFunction_New(IntPtr ml, IntPtr self);

	[DllImport("python22", CallingConvention=CallingConvention.Cdecl,
		    ExactSpelling=true, CharSet=CharSet.Ansi)]
	internal unsafe static extern IntPtr
	PyCFunction_Call(IntPtr func, IntPtr args, IntPtr kw);


	[DllImport("python22", CallingConvention=CallingConvention.Cdecl,
		    ExactSpelling=true, CharSet=CharSet.Ansi)]
	internal unsafe static extern IntPtr
	PyClass_New(IntPtr bases, IntPtr dict, IntPtr name);

	[DllImport("python22", CallingConvention=CallingConvention.Cdecl,
		    ExactSpelling=true, CharSet=CharSet.Ansi)]
	internal unsafe static extern IntPtr
	PyInstance_New(IntPtr cls, IntPtr args, IntPtr kw);



	[DllImport("python22", CallingConvention=CallingConvention.Cdecl,
		    ExactSpelling=true, CharSet=CharSet.Ansi)]
	internal unsafe static extern IntPtr
	PyMem_Malloc(int size);

	[DllImport("python22", CallingConvention=CallingConvention.Cdecl,
		    ExactSpelling=true, CharSet=CharSet.Ansi)]
	internal unsafe static extern IntPtr
	PyMem_Realloc(IntPtr ptr, int size);

	[DllImport("python22", CallingConvention=CallingConvention.Cdecl,
		    ExactSpelling=true, CharSet=CharSet.Ansi)]
	internal unsafe static extern void
	PyMem_Free(IntPtr ptr);

	[DllImport("python22", CallingConvention=CallingConvention.Cdecl,
		    ExactSpelling=true, CharSet=CharSet.Ansi)]
	internal unsafe static extern IntPtr
	PyMethod_New(IntPtr func, IntPtr self, IntPtr cls);




	// Exception APIs

	[DllImport("python22", CallingConvention=CallingConvention.Cdecl,
		    ExactSpelling=true, CharSet=CharSet.Ansi)]
	internal unsafe static extern void
	PyErr_SetString(IntPtr ob, string message);

	[DllImport("python22", CallingConvention=CallingConvention.Cdecl,
		    ExactSpelling=true, CharSet=CharSet.Ansi)]
	internal unsafe static extern void
	PyErr_SetObject(IntPtr ob, IntPtr message);

	[DllImport("python22", CallingConvention=CallingConvention.Cdecl,
		    ExactSpelling=true, CharSet=CharSet.Ansi)]
	internal unsafe static extern IntPtr
	PyErr_SetFromErrno(IntPtr ob);

	[DllImport("python22", CallingConvention=CallingConvention.Cdecl,
		    ExactSpelling=true, CharSet=CharSet.Ansi)]
	internal unsafe static extern void
	PyErr_SetNone(IntPtr ob);

	[DllImport("python22", CallingConvention=CallingConvention.Cdecl,
		    ExactSpelling=true, CharSet=CharSet.Ansi)]
	internal unsafe static extern int
	PyErr_ExceptionMatches(IntPtr exception);

	[DllImport("python22", CallingConvention=CallingConvention.Cdecl,
		    ExactSpelling=true, CharSet=CharSet.Ansi)]
	internal unsafe static extern int
	PyErr_GivenExceptionMatches(IntPtr ob, IntPtr val);

	[DllImport("python22", CallingConvention=CallingConvention.Cdecl,
		    ExactSpelling=true, CharSet=CharSet.Ansi)]
        internal unsafe static extern void
	PyErr_NormalizeException(IntPtr ob, IntPtr val, IntPtr tb);

	[DllImport("python22", CallingConvention=CallingConvention.Cdecl,
		    ExactSpelling=true, CharSet=CharSet.Ansi)]
	internal unsafe static extern int
	PyErr_Occurred();

	[DllImport("python22", CallingConvention=CallingConvention.Cdecl,
		    ExactSpelling=true, CharSet=CharSet.Ansi)]
        internal unsafe static extern void
	PyErr_Fetch(ref IntPtr ob, ref IntPtr val, ref IntPtr tb);

	[DllImport("python22", CallingConvention=CallingConvention.Cdecl,
		    ExactSpelling=true, CharSet=CharSet.Ansi)]
        internal unsafe static extern void
	PyErr_Restore(IntPtr ob, IntPtr val, IntPtr tb);

	[DllImport("python22", CallingConvention=CallingConvention.Cdecl,
		    ExactSpelling=true, CharSet=CharSet.Ansi)]
        internal unsafe static extern void
	PyErr_Clear();

    }


}
