// PyDictTests.cs - Unit tests for Python dict wrappers
//
// Author:  Brian Lloyd <brian@zope.com>
//
// (c) 2002 Brian Lloyd

using NUnit.Framework;
using Python.Runtime;
using System;

namespace Python.Runtime.Tests {


public class PyDictTests : TestCase {

    public PyDictTests(string name) : base(name) {}
    
    public static ITest Suite {
	get {
	    return new TestSuite(typeof(PyDictTests));
	}
    }

    protected override void SetUp() {
	PythonInterpreter.Initialize();
    }

    protected override void TearDown() {	
	PythonInterpreter.Finalize();
    }


    // --------------------------------------------------------------------
    // Test cases
    // --------------------------------------------------------------------

    public PyDict GetTestData() {
	PyDict dict = new PyDict();
	PyString value;

	for (int i = 0; i < 10; i++) {
    	     value = new PyString("item " + i.ToString());
	     dict.SetItem(value, value);
	}
	return dict;
    }

    public void TestPyDictCreate() {
	PyDict dict = new PyDict();
	Assert("Refcount", dict.RefCount == 1);
	Assert("Length", dict.Length == 0);
    }

    public void TestPyDictDispose() {
	PyDict dict = new PyDict();
	Assert("Handle 1", dict.Handle != IntPtr.Zero);
	Assert("Refcount 1", dict.RefCount == 1);
	dict.Dispose();
	Assert("Handle 2", dict.Handle == IntPtr.Zero);
	Assert("Refcount 2", dict.RefCount == 0);
    }

    public void TestPyDictIndexers() {
	PyDict dict = GetTestData();

	Assert("Length 1", dict.Length == 10);

	Assert("GetItem 1", dict.GetItem("item 1").ToString() == "item 1");
	Assert("GetItem 2", dict.GetItem("item 2").ToString() == "item 2");
	Assert("GetItem 3", dict.GetItem("item 3").ToString() == "item 3");

	dict.Clear();
	Assert("Clear Len", dict.Length == 0);

	dict["item 1"] = new PyString("item 1");
	dict["item 2"] = new PyString("item 2");
	dict["item 3"] = new PyString("item 3");

	Assert("Length 2", dict.Length == 3);

	Assert("Indexer 1", dict["item 1"].ToString() == "item 1");
	Assert("Indexer 2", dict["item 2"].ToString() == "item 2");
	Assert("Indexer 3", dict["item 3"].ToString() == "item 3");

	dict.DelItem("item 1");
	Assert("Length 3", dict.Length == 2);

	try {
 	    PyObject item = dict["missing"];
	    Fail("Expected KeyError");
	}
	catch (PythonException e) {
	    Assert(typeof(PythonException) == e.GetType());
	}
    }

    public void TestPyDictHasKey() {
	PyDict dict = GetTestData();
	Assert("HasKey 1", dict.HasKey(new PyString("item 1")));
	Assert("HasKey 2", dict.HasKey("item 1"));
	Assert("HasKey 3", !(dict.HasKey(new PyString("item X"))));
	Assert("HasKey 4", !(dict.HasKey("item X")));
    }

    public void TestPyDictKeys() {
	PyDict dict = GetTestData();
	PyList keys = (PyList)dict.Keys();
	Assert("Length", keys.Length == 10);
	Assert("Check 1", keys.Contains(new PyString("item 1")));
	Assert("Check 2", keys.Contains(new PyString("item 9")));
    }

    public void TestPyDictValues() {
	PyDict dict = GetTestData();
	PyList values = (PyList)dict.Values();
	Assert("Length", values.Length == 10);
	Assert("Check 1", values.Contains(new PyString("item 1")));
	Assert("Check 2", values.Contains(new PyString("item 9")));
    }

    public void TestPyDictItems() {
	PyDict dict = GetTestData();
	PyList items = (PyList)dict.Items();
	Assert("Length 1", items.Length == 10);

	for (int i = 0; i < 10; i++) {
	    Assert("Length 2", items[i].Length == 2);
	}
    }

    public void TestPyDictCopy() {
	PyDict dict = GetTestData();
	PyDict copy = dict.Copy();
	Assert("Handles", dict.Handle != copy.Handle);
	Assert("Length 1", dict.Length == 10);
	Assert("Length 2", copy.Length == 10);
    }

    public void TestPyDictUpdate() {
	PyDict dict = GetTestData();
	PyDict data = new PyDict();

	data["item 1"] = new PyString("item 1");
	data["item 12"] = new PyString("item 12");
	data["item 13"] = new PyString("item 13");
	dict.Update(data);

	Assert("Length", dict.Length == 12);
	Assert("HasKey", dict.HasKey("item 12"));
    }

    public void TestPyDictClear() {
	PyDict dict = GetTestData();
	Assert("Length 1", dict.Length == 10);
	dict.Clear();
	Assert("Length 2", dict.Length == 0);
    }


}


} // Python.Runtime.Tests
