// PyListTests.cs - Unit tests for Python list wrappers
//
// Author:  Brian Lloyd <brian@zope.com>
//
// (c) 2002 Brian Lloyd

using NUnit.Framework;
using Python.Runtime;
using System;

namespace Python.Runtime.Tests {


public class PyListTests : TestCase {

    public PyListTests(string name) : base(name) {}
    
    public static ITest Suite {
	get {
	    return new TestSuite(typeof(PyListTests));
	}
    }

    protected override void SetUp() {
	PythonInterpreter.Initialize();
    }

    protected override void TearDown() {	
	PythonInterpreter.Finalize();
    }

    // --------------------------------------------------------------------
    // Test cases
    // --------------------------------------------------------------------

    public void TestPyListCreate() {
	PyList list = new PyList();
	Assert("Initial refcount", list.RefCount == 1);
	Assert("Initial length", list.Length == 0);
    }

    public void TestPyListFromPyObjectArray() {
        PyString[] items = new PyString[10];
	for (int i = 0; i < 10; i++) {
    	    items[i] = new PyString("Item " + i.ToString());
	}
	PyList list = new PyList(items);
	Assert("Initial refcount", list.RefCount == 1);
	Assert("Initial length", list.Length == 10);
    }

    public void TestPyListDispose() {
	PyList list1 = new PyList();
	Assert("Handle is null", list1.Handle != IntPtr.Zero);
	Assert("Initial refcount", list1.RefCount == 1);
	list1.Dispose();
	Assert("Handle not null", list1.Handle == IntPtr.Zero);
	Assert("Final refcount", list1.RefCount == 0);
    }

    public void TestPyListGetItem() {
	PyList list = new PyList();
	for (int i = 0; i < 10; i++) {
    	    PyString item = new PyString("Item " + i.ToString());
	    list.Append(item);
	}

	Assert("GetItem 0", list.GetItem(0).ToString() == "Item 0");
	Assert("GetItem 9", list.GetItem(9).ToString() == "Item 9");
	Assert("GetItem -1", list.GetItem(-1).ToString() == "Item 9");
	Assert("Indexer 0", list[0].ToString() == "Item 0");
	Assert("Indexer 9", list[9].ToString() == "Item 9");
	Assert("Indexer -1", list[-1].ToString() == "Item 9");

	try {
 	    PyObject item = list[10];
	    Fail("Expected IndexError");
	}
	catch (PythonException e) {
	    Assert(typeof(PythonException) == e.GetType());
	}
    }

    public void TestPyListSetItem() {
	PyList list = new PyList();
	for (int i = 0; i < 10; i++) {
    	    PyString item = new PyString("Item " + i.ToString());
	    list.Append(item);
	}

	list.SetItem(5, new PyString("spam"));
	Assert("SetItem failed", list.GetItem(5).ToString() == "spam");

	list[5] = new PyString("eggs");
	Assert("Indexer failed", list[5].ToString() == "eggs");

	try {
	    list[10] = new PyString("spam");
	    Fail("Expected IndexError");
	}
	catch (PythonException e) {
	    Assert(typeof(PythonException) == e.GetType());
	}
    }

    public void TestPyListDelItem() {
	PyList list = new PyList();
	for (int i = 0; i < 10; i++) {
    	    PyString item = new PyString("Item " + i.ToString());
	    list.Append(item);
	}

	list.DelItem(0);
	Assert("Length failed", list.Length == 9);

	try {
	    list.DelItem(10);
	    Fail("Expected IndexError");
	}
	catch (PythonException e) {
	    Assert(typeof(PythonException) == e.GetType());
	}
    }

    public void TestPyListAppend() {
	PyList list = new PyList();
	for (int i = 0; i < 10; i++) {
    	    PyString item = new PyString("Item " + i.ToString());
	    list.Append(item);
	    Assert("Refcount failed", item.RefCount == 2);
	}
	Assert("Length failed", list.Length == 10);
    }

    public void TestPyListInsert() {
	PyList list = new PyList();
	for (int i = 0; i < 10; i++) {
    	    PyString item = new PyString("Item " + i.ToString());
	    list.Insert(0, item);
	    Assert("Refcount failed", item.RefCount == 2);
	}
	Assert("Length failed", list.Length == 10);
    }

    public void TestPyListReverse() {
	PyList list = new PyList();
	for (int i = 0; i < 10; i++) {
    	    PyString item = new PyString("Item " + i.ToString());
	    list.Append(item);
	}
	list.Reverse();
	Assert("Reverse failed", list[0].ToString() == "Item 9");
    }

    public void TestPyListIndex() {
	PyList list = new PyList();
	PyString item1 = new PyString("Item 1");
	PyString item2 = new PyString("Item 2");
	list.Append(item1);
	list.Append(item2);
	Assert("Index 0", list.Index(item1) == 0);
	Assert("Index 1", list.Index(item2) == 1);
    }

    public void TestPyListContains() {
	PyList list = new PyList();
	PyString item = null;
	for (int i = 0; i < 10; i++) {
    	    item = new PyString("Item " + i.ToString());
	    list.Append(item);
	}
	Assert("Contains failed", list.Contains(item) == true);
    }

    public void TestPyListRepeat() {
	PyList list = new PyList();
	for (int i = 0; i < 10; i++) {
    	    PyString item = new PyString("Item " + i.ToString());
	    list.Append(item);
	}
	Assert("Length failed", list.Length == 10);
	list = (PyList)list.Repeat(2);
	Assert("Repeat failed", list.Length == 20);
    }


}


} // Python.Runtime.Tests
