// PyTupleTests.cs - Unit tests for Python tuple wrappers
//
// Author:  Brian Lloyd <brian@zope.com>
//
// (c) 2002 Brian Lloyd

using NUnit.Framework;
using Python.Runtime;
using System;

namespace Python.Runtime.Tests {


public class PyTupleTests : TestCase {

    public PyTupleTests(string name) : base(name) {}
    
    public static ITest Suite {
	get {
	    return new TestSuite(typeof(PyTupleTests));
	}
    }

    protected override void SetUp() {
	PythonInterpreter.Initialize();
    }

    protected override void TearDown() {	
	PythonInterpreter.Finalize();
    }

    // --------------------------------------------------------------------
    // Test cases
    // --------------------------------------------------------------------

    private PyObject[] GetTestData() {
        PyString[] items = new PyString[10];
	for (int i = 0; i < 10; i++) {
    	    items[i] = new PyString("Item " + i.ToString());
	}
	return items;
    }

    public void TestPyTupleCreate() {
	PyTuple pyobj = new PyTuple();
	Assert("Refcount", pyobj.RefCount > 0);
	Assert("Length", pyobj.Length == 0);
    }

    public void TestPyTupleFromPyObjectArray() {
	PyObject[] items = GetTestData();
	PyTuple pyobj = new PyTuple(items);
	Assert("Refcount", pyobj.RefCount == 1);
	Assert("Length", pyobj.Length == 10);
    }

    public void TestPyTupleDispose() {
	PyTuple pyobj = new PyTuple();
	Assert("Handle 1", pyobj.Handle != IntPtr.Zero);
	Assert("Refcount 1", pyobj.RefCount > 0);
	pyobj.Dispose();
	Assert("Handle 2", pyobj.Handle == IntPtr.Zero);
	Assert("Refcount 2", pyobj.RefCount == 0);
    }

    public void TestPyTupleGetItem() {
	PyObject[] items = GetTestData();
	PyTuple pyobj = new PyTuple(items);

	Assert("GetItem 0", pyobj.GetItem(0).ToString() == "Item 0");
	Assert("GetItem 2", pyobj.GetItem(9).ToString() == "Item 9");
	Assert("GetItem 3", pyobj.GetItem(-1).ToString() == "Item 9");
	Assert("Indexer 1", pyobj[0].ToString() == "Item 0");
	Assert("Indexer 2", pyobj[9].ToString() == "Item 9");
	Assert("Indexer 3", pyobj[-1].ToString() == "Item 9");

	try {
 	    PyObject item = pyobj[10];
	    Fail("Expected IndexError");
	}
	catch (PythonException e) {
	    Assert(typeof(PythonException) == e.GetType());
	}
    }

    public void TestPyTupleSetItem() {
	PyObject[] items = GetTestData();
	PyTuple pyobj = new PyTuple(items);

	pyobj.SetItem(0, new PyString("spam"));
	Assert("SetItem failed", pyobj[0].ToString() == "spam");

	pyobj[0] = new PyString("eggs");
	Assert("Indexer failed", pyobj[0].ToString() == "eggs");
    }

    public void TestPyTupleDelItem() {
	PyObject[] items = GetTestData();
	PyTuple pyobj = new PyTuple(items);

	try {
	    pyobj.DelItem(0);
	    Fail("Expected TypeError");
	}
	catch (PythonException e) {
	    Assert(typeof(PythonException) == e.GetType());
	}
    }

    public void TestPyTupleIndex() {
	PyObject[] items = GetTestData();
	PyTuple pyobj = new PyTuple(items);

	Assert("Index 1", pyobj.Index(items[0]) == 0);
	Assert("Index 2", pyobj.Index(items[1]) == 1);
    }

    public void TestPyTupleContains() {
	PyObject[] items = GetTestData();
	PyTuple pyobj = new PyTuple(items);

	for (int i = 0; i < items.Length; i++) {
	    Assert("Contains", pyobj.Contains(items[i]));
	}
    }

    public void TestPyTupleRepeat() {
	PyObject[] items = GetTestData();
	PyTuple pyobj = new PyTuple(items);

	Assert("Length 1", pyobj.Length == 10);
	pyobj = (PyTuple)pyobj.Repeat(2);
	Assert("Length 2", pyobj.Length == 20);
    }


}


} // Python.Runtime.Tests
